import subprocess
from unittest import mock

import pytest

from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.linux.system import LinuxSystemPackageCommand


@pytest.fixture
def package_command(monkeypatch, first_app, tmp_path):
    command = LinuxSystemPackageCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.home_path = tmp_path / "home"

    # Run outside docker for these tests.
    command.target_image = None

    # Mock the detection of system python.
    command.verify_system_python = mock.MagicMock()
    command.verify_system_packages = mock.MagicMock()

    # Mock the packaging tools.
    command._verify_packaging_tools = mock.MagicMock()

    return command


def test_formats(package_command):
    """The supported packaging formats are as expected."""
    assert package_command.packaging_formats == ["deb", "rpm", "pkg", "system"]


@pytest.mark.parametrize(
    "format, vendor, codename, revision, filename",
    [
        ["rpm", "rhel", "9", None, "first-app-0.0.1-1.el9.wonky.rpm"],
        ["rpm", "rhel", "9", 5, "first-app-0.0.1-5.el9.wonky.rpm"],
        ["rpm", "fedora", "37", None, "first-app-0.0.1-1.fc37.wonky.rpm"],
        [
            "deb",
            "debian",
            "bullseye",
            None,
            "first-app_0.0.1-1~debian-bullseye_wonky.deb",
        ],
        ["deb", "debian", "bullseye", 5, "first-app_0.0.1-5~debian-bullseye_wonky.deb"],
        ["deb", "ubuntu", "jammy", None, "first-app_0.0.1-1~ubuntu-jammy_wonky.deb"],
        [
            "deb",
            "linuxmint",
            "vera",
            None,
            "first-app_0.0.1-1~linuxmint-vera_wonky.deb",
        ],
        ["pkg", "arch", "rolling", None, "first-app-0.0.1-1-wonky.pkg.tar.zst"],
        ["pkg", "manjaro", "rolling", None, "first-app-0.0.1-1-wonky.pkg.tar.zst"],
    ],
)
def test_distribution_path(
    package_command,
    first_app,
    format,
    vendor,
    codename,
    revision,
    filename,
    tmp_path,
):
    first_app.packaging_format = format
    first_app.target_vendor = vendor
    first_app.target_codename = codename

    # Mock return value for ABI from packaging system
    package_command.tools[first_app].app_context = mock.MagicMock(spec_set=Subprocess)
    package_command.tools[first_app].app_context.check_output = mock.MagicMock(
        return_value="wonky"
    )

    if revision:
        first_app.revision = revision

    assert (
        package_command.distribution_path(first_app)
        == tmp_path / "base_path/dist" / filename
    )

    # Confirm ABI was requested from build env
    package_command.tools[first_app].app_context.check_output.assert_called_with(
        {
            "deb": ["dpkg", "--print-architecture"],
            "rpm": ["rpm", "--eval", "%_target_cpu"],
            "pkg": ["pacman-conf", "Architecture"],
        }[format]
    )


@pytest.mark.parametrize("format", ["rpm", "deb", "pkg"])
def test_build_env_abi_failure(package_command, first_app, format):
    """If the subprocess to get the build ABI fails, an error is raised."""
    first_app.packaging_format = format

    # Mock return value for ABI from packaging system
    package_command.tools[first_app].app_context = mock.MagicMock(spec_set=Subprocess)
    package_command.tools[first_app].app_context.check_output = mock.MagicMock(
        side_effect=subprocess.CalledProcessError(returncode=1, cmd="pkg -arch")
    )

    with pytest.raises(
        BriefcaseCommandError,
        match="Failed to determine build environment's ABI for packaging.",
    ):
        getattr(package_command, f"{format}_abi")(first_app)


@pytest.mark.parametrize(
    "base_vendor, input_format, output_format",
    [
        # System packaging maps to known formats
        ("debian", "system", "deb"),
        ("rhel", "system", "rpm"),
        ("arch", "system", "pkg"),
        # Explicit output format is preserved
        ("debian", "deb", "deb"),
        ("redhat", "rpm", "rpm"),
        ("arch", "pkg", "pkg"),
        # This is technically possible, but probably ill-advised
        ("debian", "rpm", "rpm"),
        # Unknown base vendor, but explicit packaging format
        (None, "deb", "deb"),
        (None, "rpm", "rpm"),
        (None, "pkg", "pkg"),
    ],
)
def test_adjust_packaging_format(
    package_command,
    first_app,
    base_vendor,
    input_format,
    output_format,
):
    """The packaging format can be adjusted based on host system knowledge."""
    first_app.target_vendor_base = base_vendor
    first_app.packaging_format = input_format

    package_command.verify_app_tools(first_app)

    assert first_app.packaging_format == output_format


def test_unknown_packaging_format(package_command, first_app):
    """An unknown packaging format raises an error."""
    first_app.target_vendor_base = None
    first_app.packaging_format = "system"

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Briefcase doesn't know the system packaging format for somevendor.",
    ):
        package_command.verify_app_tools(first_app)


def test_package_deb_app(package_command, first_app):
    """A debian app can be packaged."""
    # Set the packaging format
    first_app.packaging_format = "deb"

    # Mock the actual packaging call
    package_command._package_deb = mock.MagicMock()

    # Package the app
    package_command.package_app(first_app)

    # Assert the right backend was called.
    package_command._package_deb.assert_called_once_with(first_app)


def test_package_rpm_app(package_command, first_app):
    """A Red Hat app can be packaged."""
    # Set the packaging format
    first_app.packaging_format = "rpm"

    # Mock the actual packaging call
    package_command._package_rpm = mock.MagicMock()

    # Package the app
    package_command.package_app(first_app)

    # Assert the right backend was called.
    package_command._package_rpm.assert_called_once_with(first_app)


def test_package_pkg_app(package_command, first_app):
    """An Arch app can be packaged."""
    # Set the packaging format
    first_app.packaging_format = "pkg"

    # Mock the actual packaging call
    package_command._package_pkg = mock.MagicMock()

    # Package the app
    package_command.package_app(first_app)

    # Assert the right backend was called.
    package_command._package_pkg.assert_called_once_with(first_app)


def test_package_unknown_format(package_command, first_app):
    """Unknown/unsupported packaging formats raise an error."""
    # Set the packaging format
    first_app.packaging_format = "unknown"

    # Mock the actual packaging call
    package_command._package_deb = mock.MagicMock()

    # Package the app
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Briefcase doesn't currently know how to build system packages in UNKNOWN format.",
    ):
        package_command.package_app(first_app)



from io import StringIO

import pytest

from briefcase.console import Console, Log
from briefcase.platforms.web.static import StaticWebBuildCommand

from ....utils import create_wheel


@pytest.fixture
def build_command(tmp_path):
    return StaticWebBuildCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


def test_process_wheel(build_command, tmp_path):
    """A wheel can be processed to have CSS content extracted."""

    # Create a wheel with some content.
    wheel_filename = create_wheel(
        tmp_path,
        extra_content=[
            # Two CSS files
            (
                "dummy/static/first.css",
                "span {\n  font-color: red;\n  font-size: larger\n}\n",
            ),
            ("dummy/static/second.css", "div {\n  padding: 10px\n}\n"),
            ("dummy/static/deep/third.css", "p {\n  color: red\n}\n"),
            # Content in the static file that isn't CSS
            ("dummy/static/explosions.js", "alert('boom!');"),
            # CSS in a location that isn't the static folder.
            ("dummy/other.css", "div.other {\n  margin: 10px\n}\n"),
            ("lost.css", "div.lost {\n  margin: 10px\n}\n"),
        ],
    )

    # Create a dummy css file
    css_file = StringIO()

    build_command._process_wheel(wheel_filename, css_file=css_file)

    assert (
        css_file.getvalue()
        == "\n".join(
            [
                "",
                "/*******************************************************",
                " * dummy 1.2.3::first.css",
                " *******************************************************/",
                "",
                "span {",
                "  font-color: red;",
                "  font-size: larger",
                "}",
                "",
                "/*******************************************************",
                " * dummy 1.2.3::second.css",
                " *******************************************************/",
                "",
                "div {",
                "  padding: 10px",
                "}",
                "",
                "/*******************************************************",
                " * dummy 1.2.3::deep/third.css",
                " *******************************************************/",
                "",
                "p {",
                "  color: red",
                "}",
            ]
        )
        + "\n"
    )


def test_process_wheel_no_content(build_command, tmp_path):
    """A wheel with no resources can be processed."""

    # Create a wheel with some content, but nothing static to be collected
    wheel_filename = create_wheel(
        tmp_path,
        extra_content=[
            # Content in the static file that isn't CSS
            ("dummy/static/explosions.js", "alert('boom!');"),
            # CSS in a location that isn't the static folder.
            ("dummy/other.css", "div.other {\n  margin: 10px\n}\n"),
            ("lost.css", "div.lost {\n  margin: 10px\n}\n"),
        ],
    )

    # Create a dummy css file
    css_file = StringIO()

    build_command._process_wheel(wheel_filename, css_file=css_file)

    assert css_file.getvalue() == ""



import pytest

from briefcase.commands.new import titlecase


@pytest.mark.parametrize(
    "raw, converted",
    [
        ("Hello", "Hello"),
        ("hello", "Hello"),
        ("hello world", "Hello World"),
        ("hello world's fair", "Hello World's Fair"),
        ("world of wonder", "World of Wonder"),
        ("hunt the wumpus", "Hunt the Wumpus"),
        ("scott pilgrim vs the world", "Scott Pilgrim vs the World"),
        # Specific examples that we know are problematic
        ("author's email", "Author's Email"),
        ("project URL", "Project URL"),
    ],
)
def test_titlecase(raw, converted):
    """Test that a string can be capitalized."""
    assert titlecase(raw) == converted



import subprocess
import sys
from unittest import mock

import pytest


def test_dev_run(dev_command, first_app, tmp_path):
    """The app can be run in dev mode."""
    dev_command._stream_app_logs = mock.MagicMock()
    app_popen = mock.MagicMock()
    dev_command.tools.subprocess.Popen.return_value = app_popen

    dev_command.run_dev_app(
        first_app,
        env={"a": 1, "b": 2, "c": 3},
        test_mode=False,
        passthrough=[],
    )

    dev_command.tools.subprocess.Popen.assert_called_once_with(
        [
            sys.executable,
            "-c",
            (
                "import runpy, sys;"
                "sys.path.pop(0);"
                "sys.argv.extend([]);"
                'runpy.run_module("first", run_name="__main__", alter_sys=True)'
            ),
        ],
        env={
            "a": 1,
            "b": 2,
            "c": 3,
            "PYTHONUNBUFFERED": "1",
            "PYTHONDEVMODE": "1",
            "PYTHONUTF8": "1",
        },
        cwd=dev_command.tools.home_path,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        encoding="UTF-8",
    )
    dev_command._stream_app_logs.assert_called_once_with(
        first_app,
        popen=app_popen,
        test_mode=False,
        clean_output=False,
    )


def test_dev_run_with_args(dev_command, first_app, tmp_path):
    "The app can be run in dev mode with arguments"
    dev_command._stream_app_logs = mock.MagicMock()
    app_popen = mock.MagicMock()
    dev_command.tools.subprocess.Popen.return_value = app_popen

    dev_command.run_dev_app(
        first_app,
        env={"a": 1, "b": 2, "c": 3},
        test_mode=False,
        passthrough=["foo", "bar", "--whiz"],
    )

    dev_command.tools.subprocess.Popen.assert_called_once_with(
        [
            sys.executable,
            "-c",
            (
                "import runpy, sys;"
                "sys.path.pop(0);"
                "sys.argv.extend(['foo', 'bar', '--whiz']);"
                'runpy.run_module("first", run_name="__main__", alter_sys=True)'
            ),
        ],
        env={
            "a": 1,
            "b": 2,
            "c": 3,
            "PYTHONUNBUFFERED": "1",
            "PYTHONDEVMODE": "1",
            "PYTHONUTF8": "1",
        },
        cwd=dev_command.tools.home_path,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        encoding="UTF-8",
    )
    dev_command._stream_app_logs.assert_called_once_with(
        first_app,
        popen=app_popen,
        test_mode=False,
        clean_output=False,
    )


@pytest.mark.parametrize("is_console_app", [True, False])
def test_dev_test_mode(dev_command, first_app, is_console_app, tmp_path):
    """The test suite can be run in development mode."""
    # Test mode is the same regardless of whether it's a console app or not.
    first_app.console_app = is_console_app

    dev_command._stream_app_logs = mock.MagicMock()
    app_popen = mock.MagicMock()
    dev_command.tools.subprocess.Popen.return_value = app_popen

    dev_command.run_dev_app(
        first_app,
        env={"a": 1, "b": 2, "c": 3},
        test_mode=True,
        passthrough=[],
    )

    dev_command.tools.subprocess.Popen.assert_called_once_with(
        [
            sys.executable,
            "-c",
            (
                "import runpy, sys;"
                "sys.path.pop(0);"
                "sys.argv.extend([]);"
                'runpy.run_module("tests.first", run_name="__main__", alter_sys=True)'
            ),
        ],
        env={
            "a": 1,
            "b": 2,
            "c": 3,
            "PYTHONUNBUFFERED": "1",
            "PYTHONDEVMODE": "1",
            "PYTHONUTF8": "1",
        },
        cwd=dev_command.tools.home_path,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        encoding="UTF-8",
    )
    dev_command._stream_app_logs.assert_called_once_with(
        first_app,
        popen=app_popen,
        test_mode=True,
        clean_output=False,
    )


@pytest.mark.parametrize("is_console_app", [True, False])
def test_dev_test_mode_with_args(dev_command, first_app, is_console_app, tmp_path):
    """The test suite can be run in development mode with args."""
    # Test mode is the same regardless of whether it's a console app or not.
    first_app.console_app = is_console_app

    dev_command._stream_app_logs = mock.MagicMock()
    app_popen = mock.MagicMock()
    dev_command.tools.subprocess.Popen.return_value = app_popen

    dev_command.run_dev_app(
        first_app,
        env={"a": 1, "b": 2, "c": 3},
        test_mode=True,
        passthrough=["foo", "bar", "--whiz"],
    )

    dev_command.tools.subprocess.Popen.assert_called_once_with(
        [
            sys.executable,
            "-c",
            (
                "import runpy, sys;"
                "sys.path.pop(0);"
                "sys.argv.extend(['foo', 'bar', '--whiz']);"
                'runpy.run_module("tests.first", run_name="__main__", alter_sys=True)'
            ),
        ],
        env={
            "a": 1,
            "b": 2,
            "c": 3,
            "PYTHONUNBUFFERED": "1",
            "PYTHONDEVMODE": "1",
            "PYTHONUTF8": "1",
        },
        cwd=dev_command.tools.home_path,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        encoding="UTF-8",
    )
    dev_command._stream_app_logs.assert_called_once_with(
        first_app,
        popen=app_popen,
        test_mode=True,
        clean_output=False,
    )


def test_dev_run_console(dev_command, first_app, tmp_path):
    """A console app can be run in dev mode."""
    # Modify the app to be a console app
    first_app.console_app = True

    dev_command._stream_app_logs = mock.MagicMock()

    dev_command.run_dev_app(
        first_app,
        env={"a": 1, "b": 2, "c": 3},
        test_mode=False,
        passthrough=[],
    )

    dev_command.tools.subprocess.run.assert_called_once_with(
        [
            sys.executable,
            "-c",
            (
                "import runpy, sys;"
                "sys.path.pop(0);"
                "sys.argv.extend([]);"
                'runpy.run_module("first", run_name="__main__", alter_sys=True)'
            ),
        ],
        env={
            "a": 1,
            "b": 2,
            "c": 3,
            "PYTHONUNBUFFERED": "1",
            "PYTHONDEVMODE": "1",
            "PYTHONUTF8": "1",
        },
        cwd=dev_command.tools.home_path,
        bufsize=1,
        encoding="UTF-8",
        stream_output=False,
    )

    # There's no log streaming
    dev_command._stream_app_logs.assert_not_called()


def test_dev_run_console_with_args(dev_command, first_app, tmp_path):
    "The console app can be run in dev mode with arguments"
    # Modify the app to be a console app
    first_app.console_app = True

    dev_command._stream_app_logs = mock.MagicMock()

    dev_command.run_dev_app(
        first_app,
        env={"a": 1, "b": 2, "c": 3},
        test_mode=False,
        passthrough=["foo", "bar", "--whiz"],
    )

    dev_command.tools.subprocess.run.assert_called_once_with(
        [
            sys.executable,
            "-c",
            (
                "import runpy, sys;"
                "sys.path.pop(0);"
                "sys.argv.extend(['foo', 'bar', '--whiz']);"
                'runpy.run_module("first", run_name="__main__", alter_sys=True)'
            ),
        ],
        env={
            "a": 1,
            "b": 2,
            "c": 3,
            "PYTHONUNBUFFERED": "1",
            "PYTHONDEVMODE": "1",
            "PYTHONUTF8": "1",
        },
        cwd=dev_command.tools.home_path,
        bufsize=1,
        encoding="UTF-8",
        stream_output=False,
    )

    # No attempt to stream logs
    dev_command._stream_app_logs.assert_not_called()



from briefcase.integrations.xcode import Xcode

DEFAULT_OUTPUT_FORMAT = "Xcode"


class iOSMixin:
    platform = "iOS"
    supported_host_os = {"Darwin"}
    supported_host_os_reason = (
        "iOS applications require Xcode, which is only available on macOS."
    )
    # 0.3.16 introduced new-style dylib support.
    platform_target_version = "0.3.16"

    def verify_tools(self):
        Xcode.verify(self.tools, min_version=(13, 0, 0))

        # Verify superclass tools *after* xcode. This ensures we get the
        # git check *after* the xcode check.
        super().verify_tools()



import subprocess
from unittest.mock import ANY, MagicMock

import pytest

from briefcase.console import Console, Log, LogLevel
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.macOS.xcode import macOSXcodeBuildCommand


@pytest.fixture
def build_command(tmp_path):
    return macOSXcodeBuildCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


@pytest.mark.parametrize("tool_debug_mode", (True, False))
def test_build_app(build_command, first_app_generated, tool_debug_mode, tmp_path):
    """An macOS App can be built."""
    # Enable verbose tool logging
    if tool_debug_mode:
        build_command.tools.logger.verbosity = LogLevel.DEEP_DEBUG

    build_command.tools.subprocess = MagicMock(spec_set=Subprocess)
    build_command.build_app(first_app_generated, test_mode=False)

    build_command.tools.subprocess.run.assert_called_with(
        [
            "xcodebuild",
            "-project",
            tmp_path
            / "base_path"
            / "build"
            / "first-app"
            / "macos"
            / "xcode"
            / "First App.xcodeproj",
            "-verbose" if tool_debug_mode else "-quiet",
            "-configuration",
            "Release",
            "build",
        ],
        check=True,
        filter_func=None if tool_debug_mode else ANY,
    )


def test_build_app_failed(build_command, first_app_generated, tmp_path):
    """If xcodebuild fails, an error is raised."""
    # The subprocess.run() call will raise an error
    build_command.tools.subprocess = MagicMock(spec_set=Subprocess)
    build_command.tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        cmd=["xcodebuild", "..."],
        returncode=1,
    )

    with pytest.raises(BriefcaseCommandError):
        build_command.build_app(first_app_generated, test_mode=False)

    build_command.tools.subprocess.run.assert_called_with(
        [
            "xcodebuild",
            "-project",
            tmp_path
            / "base_path"
            / "build"
            / "first-app"
            / "macos"
            / "xcode"
            / "First App.xcodeproj",
            "-quiet",
            "-configuration",
            "Release",
            "build",
        ],
        check=True,
        filter_func=ANY,
    )



from unittest.mock import ANY, MagicMock

from ...utils import NoMatchString, PartialMatchString


def test_no_test_dir(convert_command, monkeypatch):
    mock_input_text = MagicMock()
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)

    convert_command.input_test_source_dir("app_name", None)
    intro_content = "\n\nBased on your project's folder structure, we believe 'test' might be your test directory"
    mock_input_text.assert_called_once_with(
        intro=NoMatchString(intro_content),
        variable="test source directory",
        default="tests",
        validator=ANY,
        override_value=None,
    )


def test_test_dir(convert_command, monkeypatch):
    mock_input_text = MagicMock()
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)

    (convert_command.base_path / "test").mkdir()
    convert_command.input_test_source_dir("app_name", None)

    intro_content = "\n\nBased on your project's folder structure, we believe 'test' might be your test directory"
    mock_input_text.assert_called_once_with(
        intro=PartialMatchString(intro_content),
        variable="test source directory",
        default="test",
        validator=ANY,
        override_value=None,
    )


def test_tests_dir(convert_command, monkeypatch):
    mock_input_text = MagicMock()
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)

    (convert_command.base_path / "tests").mkdir()
    convert_command.input_test_source_dir("app_name", None)

    intro_content = "\n\nBased on your project's folder structure, we believe 'tests' might be your test directory"
    mock_input_text.assert_called_once_with(
        intro=PartialMatchString(intro_content),
        variable="test source directory",
        default="tests",
        validator=ANY,
        override_value=None,
    )


def test_tests_dir_is_prefered_over_test_dir(convert_command, monkeypatch):
    mock_input_text = MagicMock()
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)

    (convert_command.base_path / "tests").mkdir()
    (convert_command.base_path / "test").mkdir()
    convert_command.input_test_source_dir("app_name", None)

    intro_content = "\n\nBased on your project's folder structure, we believe 'tests' might be your test directory"
    mock_input_text.assert_called_once_with(
        intro=PartialMatchString(intro_content),
        variable="test source directory",
        default="tests",
        validator=ANY,
        override_value=None,
    )


def test_override_is_used(convert_command):
    (convert_command.base_path / "test_dir").mkdir()
    assert convert_command.input_test_source_dir("app_name", "test_dir") == "test_dir"


def test_prompted_test_source_dir(convert_command):
    """You can type in the test source dir."""
    (convert_command.base_path / "mytest").mkdir(parents=True)
    convert_command.input.values = ["mytest"]

    assert convert_command.input_test_source_dir("app_name", None) == "mytest"



from pathlib import Path
from random import randrange

from briefcase.integrations.linuxdeploy import (
    ELF_PATCH_OFFSET,
    ELF_PATCH_ORIGINAL_BYTES,
)


def create_mock_appimage(appimage_path: Path, mock_appimage_kind: str = "original"):
    """Create a mock AppImage for testing purposes.

    :param appimage_path: Path to the appimage to create.
    :param mock_appimage_kind: The kind of mock appimage to create. 'original' creates
        an unpatched mock appimage. 'patched' creates a patched mock appimage. 'corrupt'
        creates a corrupted mock appimage.
    :returns: The bytes to be patched of the created AppImage.
    """

    bytes_to_be_patched = None

    appimage_headers = {
        "original": bytes.fromhex("7f454c46020101004149020000000000"),
        "patched": bytes.fromhex("7f454c46020101000000000000000000"),
        "corrupt": bytes.fromhex("%030x" % randrange(16**30)),
        "empty": b"",
    }

    appimage_path.parent.mkdir(parents=True, exist_ok=True)
    appimage_path.touch()
    with open(appimage_path, "w+b") as mock_appimage:
        if mock_appimage_kind in appimage_headers:
            mock_appimage.write(appimage_headers[mock_appimage_kind])
        else:
            raise ValueError(f"Unknown mock_appimage_kind: {mock_appimage_kind}")
        mock_appimage.seek(ELF_PATCH_OFFSET)
        bytes_to_be_patched = mock_appimage.read(len(ELF_PATCH_ORIGINAL_BYTES))

    return bytes_to_be_patched


def side_effect_create_mock_appimage(appimage_path):
    """A test fixture side effect that creates an AppImage at the specified path."""

    def _side_effect(*args, **kwargs):
        create_mock_appimage(appimage_path=appimage_path)
        return "new-downloaded-file"

    return _side_effect


def side_effect_create_mock_tool(tool_path):
    """A test fixture side effect that creates an sh file at the specified path."""

    def _side_effect(*args, **kwargs):
        tool_path.parent.mkdir(parents=True)
        with tool_path.open("w", encoding="utf-8") as f:
            f.write("I am a complete tool")
        return "new-downloaded-file"

    return _side_effect



import datetime
import os
import subprocess
import sys
from unittest import mock

import pytest
import tomli_w

import briefcase
from briefcase.commands.create import _is_local_requirement
from briefcase.console import LogLevel
from briefcase.exceptions import BriefcaseCommandError, RequirementsInstallError
from briefcase.integrations.subprocess import Subprocess

GENERATED_DATETIME = "# Generated 2024-05-02 12:00:00.000500"


@pytest.fixture
def create_command(create_command, myapp):
    # mock subprocess app context for this app
    create_command.tools[myapp].app_context = mock.MagicMock(spec_set=Subprocess)
    return create_command


@pytest.fixture
def mock_now(monkeypatch):
    """Monkeypatch the ``datetime.now`` inside ``briefcase.commands.create``.

    When this fixture is used, the datetime is locked to 2024 May 2 @ 12:00:00:000500.
    """
    now = datetime.datetime(2024, 5, 2, 12, 0, 0, 500)
    datetime_mock = mock.MagicMock(wraps=datetime.datetime)
    datetime_mock.now.return_value = now
    monkeypatch.setattr(briefcase.commands.create, "datetime", datetime_mock)
    return now


def create_installation_artefacts(app_packages_path, packages):
    """Utility method for generating a function that will mock installation artefacts.

    Creates a function that when invoked, creates a dummy ``__init__.py``
    and ``__main__.py`` for each package named in ``packages``.

    :param app_packages_path: The pathlib object where app packages will be installed
    :param packages: A list of package names to mock.
    :returns: A function that will create files to mock the named installed packages.
    """

    def _create_installation_artefacts(*args, **kwargs):
        for package in packages:
            package_path = app_packages_path / package
            package_path.mkdir(parents=True)
            with (package_path / "__init__.py").open("w", encoding="utf-8") as f:
                f.write("")
            with (package_path / "__main__.py").open("w", encoding="utf-8") as f:
                f.write('print("I am {package}")')

    return _create_installation_artefacts


def test_bad_path_index(create_command, myapp, bundle_path, app_requirements_path):
    """If the app's path index doesn't declare a destination for requirements, an error
    is raised."""
    # Write a briefcase.toml that is missing app_packages_path and app_requirements_path
    with (bundle_path / "briefcase.toml").open("wb") as f:
        index = {
            "paths": {
                "app_path": "path/to/app",
                "support_path": "path/to/support",
            }
        }
        tomli_w.dump(index, f)

    # Set up requirements for the app
    myapp.requires = ["first", "second", "third"]

    # Install requirements
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Application path index file does not define `app_requirements_path` or `app_packages_path`",
    ):
        create_command.install_app_requirements(myapp, test_mode=False)

    # pip wasn't invoked
    create_command.tools[myapp].app_context.run.assert_not_called()

    # requirements.txt doesn't exist either
    assert not app_requirements_path.exists()

    # Original app definitions haven't changed
    assert myapp.requires == ["first", "second", "third"]
    assert myapp.test_requires is None


def test_app_packages_no_requires(
    create_command,
    myapp,
    app_packages_path,
    app_packages_path_index,
):
    """If an app has no requirements, install_app_requirements is a no-op."""
    myapp.requires = None

    create_command.install_app_requirements(myapp, test_mode=False)

    # No request was made to install requirements
    create_command.tools[myapp].app_context.run.assert_not_called()


def test_app_packages_empty_requires(
    create_command,
    myapp,
    app_packages_path,
    app_packages_path_index,
):
    """If an app has an empty requirements list, install_app_requirements is a no-op."""
    myapp.requires = []

    create_command.install_app_requirements(myapp, test_mode=False)

    # No request was made to install requirements
    create_command.tools[myapp].app_context.run.assert_not_called()


def test_app_packages_valid_requires(
    create_command,
    myapp,
    app_packages_path,
    app_packages_path_index,
):
    """If an app has a valid list of requirements, pip is invoked."""
    myapp.requires = ["first", "second==1.2.3", "third>=3.2.1"]

    create_command.install_app_requirements(myapp, test_mode=False)

    # A request was made to install requirements
    create_command.tools[myapp].app_context.run.assert_called_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--upgrade",
            "--no-user",
            f"--target={app_packages_path}",
            "first",
            "second==1.2.3",
            "third>=3.2.1",
        ],
        check=True,
        encoding="UTF-8",
    )

    # Original app definitions haven't changed
    assert myapp.requires == ["first", "second==1.2.3", "third>=3.2.1"]
    assert myapp.test_requires is None


def test_app_packages_valid_requires_no_support_package(
    create_command,
    myapp,
    app_packages_path,
    app_packages_path_index,
):
    """If the template doesn't specify a support package, the cross-platform site isn't
    specified."""
    myapp.requires = ["first", "second==1.2.3", "third>=3.2.1"]

    # Override the cache of paths to specify an app packages path, but no support package path
    create_command._briefcase_toml[myapp] = {
        "paths": {"app_packages_path": "path/to/app_packages"}
    }

    create_command.install_app_requirements(myapp, test_mode=False)

    # A request was made to install requirements
    create_command.tools[myapp].app_context.run.assert_called_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--upgrade",
            "--no-user",
            f"--target={app_packages_path}",
            "first",
            "second==1.2.3",
            "third>=3.2.1",
        ],
        check=True,
        encoding="UTF-8",
    )

    # Original app definitions haven't changed
    assert myapp.requires == ["first", "second==1.2.3", "third>=3.2.1"]
    assert myapp.test_requires is None


def test_app_packages_invalid_requires(
    create_command,
    myapp,
    app_packages_path,
    app_packages_path_index,
):
    """If an app has a valid list of requirements, pip is invoked."""
    myapp.requires = ["does-not-exist"]

    # Unfortunately, no way to tell the difference between "offline" and
    # "your requirements are invalid"; pip returns status code 1 for all
    # failures.
    create_command.tools[myapp].app_context.run.side_effect = (
        subprocess.CalledProcessError(
            cmd=["python", "-u", "-m", "pip", "..."], returncode=1
        )
    )

    with pytest.raises(RequirementsInstallError):
        create_command.install_app_requirements(myapp, test_mode=False)

    # But the request to install was still made
    create_command.tools[myapp].app_context.run.assert_called_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--upgrade",
            "--no-user",
            f"--target={app_packages_path}",
            "does-not-exist",
        ],
        check=True,
        encoding="UTF-8",
    )

    # Original app definitions haven't changed
    assert myapp.requires == ["does-not-exist"]
    assert myapp.test_requires is None


def test_app_packages_offline(
    create_command,
    myapp,
    app_packages_path,
    app_packages_path_index,
):
    """If user is offline, pip fails."""
    myapp.requires = ["first", "second", "third"]

    # Unfortunately, no way to tell the difference between "offline" and
    # "your requirements are invalid"; pip returns status code 1 for all
    # failures.
    create_command.tools[myapp].app_context.run.side_effect = (
        subprocess.CalledProcessError(
            cmd=["python", "-u", "-m", "pip", "..."], returncode=1
        )
    )

    with pytest.raises(RequirementsInstallError):
        create_command.install_app_requirements(myapp, test_mode=False)

    # But the request to install was still made
    create_command.tools[myapp].app_context.run.assert_called_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--upgrade",
            "--no-user",
            f"--target={app_packages_path}",
            "first",
            "second",
            "third",
        ],
        check=True,
        encoding="UTF-8",
    )

    # Original app definitions haven't changed
    assert myapp.requires == ["first", "second", "third"]
    assert myapp.test_requires is None


@pytest.mark.parametrize("logging_level", [LogLevel.INFO, LogLevel.DEEP_DEBUG])
def test_app_packages_install_requirements(
    create_command,
    myapp,
    app_packages_path,
    app_packages_path_index,
    logging_level,
):
    """Requirements can be installed."""
    # Configure logging level
    create_command.logger.verbosity = logging_level

    # Set up the app requirements
    myapp.requires = ["first", "second", "third"]

    # The side effect of calling pip is creating installation artefacts
    create_command.tools[myapp].app_context.run.side_effect = (
        create_installation_artefacts(app_packages_path, myapp.requires)
    )

    # Install the requirements
    create_command.install_app_requirements(myapp, test_mode=False)

    # The request to install was made
    create_command.tools[myapp].app_context.run.assert_called_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--upgrade",
            "--no-user",
            f"--target={app_packages_path}",
        ]
        + (["-vv"] if logging_level == LogLevel.DEEP_DEBUG else [])
        + ["first", "second", "third"],
        check=True,
        encoding="UTF-8",
    )

    # The new app packages have installation artefacts created
    assert (app_packages_path / "first").exists()
    assert (app_packages_path / "first/__main__.py").exists()
    assert (app_packages_path / "second").exists()
    assert (app_packages_path / "second/__main__.py").exists()
    assert (app_packages_path / "third").exists()
    assert (app_packages_path / "third/__main__.py").exists()

    # Original app definitions haven't changed
    assert myapp.requires == ["first", "second", "third"]
    assert myapp.test_requires is None


def test_app_packages_replace_existing_requirements(
    create_command,
    myapp,
    app_packages_path,
    app_packages_path_index,
):
    """If the app has already had requirements installed, they are removed first."""
    # Create some existing requirements
    create_installation_artefacts(app_packages_path, ["old", "ancient"])()

    # Set up the app requirements
    myapp.requires = ["first", "second", "third"]

    # The side effect of calling pip is creating installation artefacts
    create_command.tools[myapp].app_context.run.side_effect = (
        create_installation_artefacts(app_packages_path, myapp.requires)
    )

    # Install the requirements
    create_command.install_app_requirements(myapp, test_mode=False)

    # The request to install was still made
    create_command.tools[myapp].app_context.run.assert_called_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--upgrade",
            "--no-user",
            f"--target={app_packages_path}",
            "first",
            "second",
            "third",
        ],
        check=True,
        encoding="UTF-8",
    )

    # The new app packages have installation artefacts created
    assert (app_packages_path / "first").exists()
    assert (app_packages_path / "first/__main__.py").exists()
    assert (app_packages_path / "second").exists()
    assert (app_packages_path / "second/__main__.py").exists()
    assert (app_packages_path / "third").exists()
    assert (app_packages_path / "third/__main__.py").exists()

    # The old app packages no longer exist.
    assert not (app_packages_path / "old").exists()
    assert not (app_packages_path / "ancient").exists()

    # Original app definitions haven't changed
    assert myapp.requires == ["first", "second", "third"]
    assert myapp.test_requires is None


def test_app_requirements_no_requires(
    create_command,
    myapp,
    app_requirements_path,
    app_requirements_path_index,
):
    """If an app has no requirements, a requirements file is still written."""
    myapp.requires = None

    # Install requirements into the bundle
    create_command.install_app_requirements(myapp, test_mode=False)

    # requirements.txt doesn't exist either
    assert app_requirements_path.exists()
    with app_requirements_path.open(encoding="utf-8") as f:
        assert f.read() == ""

    # Original app definitions haven't changed
    assert myapp.requires is None
    assert myapp.test_requires is None


def test_app_requirements_empty_requires(
    create_command,
    myapp,
    app_requirements_path,
    app_requirements_path_index,
):
    """If an app has an empty requirements list, a requirements file is still
    written."""
    myapp.requires = []

    # Install requirements into the bundle
    create_command.install_app_requirements(myapp, test_mode=False)

    # requirements.txt doesn't exist either
    assert app_requirements_path.exists()
    with app_requirements_path.open(encoding="utf-8") as f:
        assert f.read() == ""

    # Original app definitions haven't changed
    assert myapp.requires == []
    assert myapp.test_requires is None


def test_app_requirements_requires(
    create_command,
    myapp,
    app_requirements_path,
    app_requirements_path_index,
    mock_now,
):
    """If an app has an empty requirements list, a requirements file is still
    written."""
    myapp.requires = ["first", "second==1.2.3", "third>=3.2.1"]

    # Install requirements into the bundle
    create_command.install_app_requirements(myapp, test_mode=False)

    # requirements.txt doesn't exist either
    assert app_requirements_path.exists()
    with app_requirements_path.open(encoding="utf-8") as f:
        assert f.read() == f"{GENERATED_DATETIME}\nfirst\nsecond==1.2.3\nthird>=3.2.1\n"

    # Original app definitions haven't changed
    assert myapp.requires == ["first", "second==1.2.3", "third>=3.2.1"]
    assert myapp.test_requires is None


@pytest.mark.parametrize(
    "altsep, requirement, expected",
    [
        (None, "asdf/xcvb", True),
        (None, "asdf>xcvb", False),
        (">", "asdf/xcvb", True),
        (">", "asdf>xcvb", True),
        (">", "asdf+xcvb", False),
    ],
)
def test__is_local_requirement_altsep_respected(
    altsep,
    requirement,
    expected,
    monkeypatch,
):
    """``os.altsep`` is included as a separator when available."""
    monkeypatch.setattr(os, "sep", "/")
    monkeypatch.setattr(os, "altsep", altsep)
    assert _is_local_requirement(requirement) is expected


def _test_app_requirements_paths(
    create_command,
    myapp,
    app_requirements_path,
    tmp_path,
    requirement,
):
    """A utility method that can be used to test expansion of a specific requirement."""
    if isinstance(requirement, tuple):
        requirement, converted = requirement
    else:
        converted = requirement
    myapp.requires = ["first", requirement, "third"]

    create_command.install_app_requirements(myapp, test_mode=False)
    with app_requirements_path.open(encoding="utf-8") as f:
        assert f.read() == (
            "\n".join(
                [
                    GENERATED_DATETIME,
                    "first",
                    converted.format(tmp_path),
                    "third",
                ]
            )
            + "\n"
        )

    # Original app definitions haven't changed
    assert myapp.requires == ["first", requirement, "third"]
    assert myapp.test_requires is None


@pytest.mark.parametrize(
    "requirement",
    [
        # Simple PyPI package references
        "my-package",
        "my-package==1.2.3",
        "my-package<=1.2.3",
        # More complex PyPI references
        "my-package[optional]<=1.2.3",
        "my-package[optional]<=1.2.3; python_version<3.7",
        # References to git packages
        "git+https://github.com/project/package",
        "git+https://github.com/project/package#egg=my-package",
        "git+https://github.com/project/package@deadbeef#egg=my-package",
        "git+https://github.com/project/package@some-branch#egg=my-package",
        # URL references to wheels
        "http://example.com/path/to/mypackage-1.2.3-py3-none-any.whl",
        "https://example.com/path/to/mypackage-1.2.3-py3-none-any.whl",
        # Zip file source installs
        "my-package @ https://example.com/path/to/1.2.3.zip",
    ],
)
def test_app_requirements_non_paths(
    create_command,
    myapp,
    app_requirements_path,
    app_requirements_path_index,
    tmp_path,
    requirement,
    mock_now,
):
    """Requirements which are not paths are left unchanged."""
    _test_app_requirements_paths(
        create_command,
        myapp,
        app_requirements_path,
        tmp_path,
        requirement,
    )


@pytest.mark.skipif(os.name != "posix", reason="Unix specific tests")
@pytest.mark.parametrize(
    "requirement",
    [
        # A reference that involves an absolute path
        "/absolute/path/to/package",
        # Relative paths.
        ("./package/inside/project", "{}/base_path/package/inside/project"),
        ("../package/outside/project", "{}/package/outside/project"),
        ("sub/package/inside/project", "{}/base_path/sub/package/inside/project"),
    ],
)
def test_app_requirements_paths_unix(
    create_command,
    myapp,
    app_requirements_path,
    app_requirements_path_index,
    tmp_path,
    requirement,
    mock_now,
):
    """Requirement paths in Unix format are expanded correctly."""
    _test_app_requirements_paths(
        create_command,
        myapp,
        app_requirements_path,
        tmp_path,
        requirement,
    )


@pytest.mark.skipif(os.name != "nt", reason="Windows specific tests")
@pytest.mark.parametrize(
    "requirement",
    [
        # A reference that involves an absolute path
        r"C:\absolute\path\to\package",
        ("C:/absolute/path/to/package", r"C:\absolute\path\to\package"),
        ("/absolute/path/to/package", r"C:\absolute\path\to\package"),
        # Relative paths using forward slash separators
        ("./package/inside/project", r"{}\base_path\package\inside\project"),
        ("../package/outside/project", r"{}\package\outside\project"),
        ("sub/package/inside/project", r"{}\base_path\sub\package\inside\project"),
        # Relative paths using backslash separators
        (r".\package\inside\project", r"{}\base_path\package\inside\project"),
        (r"..\package\outside\project", r"{}\package\outside\project"),
        (r"sub\package\inside\project", r"{}\base_path\sub\package\inside\project"),
    ],
)
def test_app_requirements_paths_windows(
    create_command,
    myapp,
    app_requirements_path,
    app_requirements_path_index,
    tmp_path,
    requirement,
    mock_now,
):
    """Requirement paths in Windows format are expanded correctly."""
    _test_app_requirements_paths(
        create_command,
        myapp,
        app_requirements_path,
        tmp_path,
        requirement,
    )


def test_app_packages_test_requires(
    create_command,
    myapp,
    app_packages_path,
    app_packages_path_index,
):
    """If an app has test requirements, they're not included unless we are in test
    mode."""
    myapp.requires = ["first", "second==1.2.3", "third>=3.2.1"]
    myapp.test_requires = ["pytest", "pytest-tldr"]

    create_command.install_app_requirements(myapp, test_mode=False)

    # A request was made to install requirements
    create_command.tools[myapp].app_context.run.assert_called_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--upgrade",
            "--no-user",
            f"--target={app_packages_path}",
            "first",
            "second==1.2.3",
            "third>=3.2.1",
        ],
        check=True,
        encoding="UTF-8",
    )

    # Original app definitions haven't changed
    assert myapp.requires == ["first", "second==1.2.3", "third>=3.2.1"]
    assert myapp.test_requires == ["pytest", "pytest-tldr"]


def test_app_packages_test_requires_test_mode(
    create_command,
    myapp,
    app_packages_path,
    app_packages_path_index,
):
    """If an app has test requirements and we're in test mode, they are installed."""
    myapp.requires = ["first", "second==1.2.3", "third>=3.2.1"]
    myapp.test_requires = ["pytest", "pytest-tldr"]

    create_command.install_app_requirements(myapp, test_mode=True)

    # A request was made to install requirements
    create_command.tools[myapp].app_context.run.assert_called_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--upgrade",
            "--no-user",
            f"--target={app_packages_path}",
            "first",
            "second==1.2.3",
            "third>=3.2.1",
            "pytest",
            "pytest-tldr",
        ],
        check=True,
        encoding="UTF-8",
    )

    # Original app definitions haven't changed
    assert myapp.requires == ["first", "second==1.2.3", "third>=3.2.1"]
    assert myapp.test_requires == ["pytest", "pytest-tldr"]


def test_app_packages_only_test_requires_test_mode(
    create_command,
    myapp,
    app_packages_path,
    app_packages_path_index,
):
    """If an app only has test requirements and we're in test mode, they are
    installed."""
    myapp.requires = None
    myapp.test_requires = ["pytest", "pytest-tldr"]

    create_command.install_app_requirements(myapp, test_mode=True)

    # A request was made to install requirements
    create_command.tools[myapp].app_context.run.assert_called_with(
        [
            sys.executable,
            "-u",
            "-X",
            "utf8",
            "-m",
            "pip",
            "install",
            "--disable-pip-version-check",
            "--no-python-version-warning",
            "--upgrade",
            "--no-user",
            f"--target={app_packages_path}",
            "pytest",
            "pytest-tldr",
        ],
        check=True,
        encoding="UTF-8",
    )

    # Original app definitions haven't changed
    assert myapp.requires is None
    assert myapp.test_requires == ["pytest", "pytest-tldr"]



import pytest

from briefcase.console import Console, Log
from briefcase.platforms.linux.system import LinuxSystemCreateCommand

from ....utils import create_file


@pytest.fixture
def create_command(tmp_path):
    return LinuxSystemCreateCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


@pytest.fixture
def first_app(first_app_config, tmp_path):
    """A fixture for the first app, rolled out on disk."""
    # Specify a system python app for a dummy vendor
    first_app_config.target_vendor = "somevendor"
    first_app_config.target_codename = "surprising"
    first_app_config.target_vendor_base = "basevendor"

    # Targeted Python version
    first_app_config.python_version_tag = "3"

    # Some project-level files.
    create_file(tmp_path / "base_path/LICENSE", "First App License")
    create_file(tmp_path / "base_path/CHANGELOG", "First App Changelog")

    # Make it look like the template has been generated
    bundle_dir = tmp_path / "base_path/build/first-app/somevendor/surprising"

    create_file(bundle_dir / "first-app.1", "First App manpage")

    lib_dir = bundle_dir / "first-app-0.0.1/usr/lib/first-app"
    (lib_dir / "app").mkdir(parents=True, exist_ok=True)
    (lib_dir / "app_packages/firstlib").mkdir(parents=True, exist_ok=True)
    (lib_dir / "app_packages/secondlib").mkdir(parents=True, exist_ok=True)

    # Create some .so files
    # An SO file with different group and world permissions
    (lib_dir / "app/support.so").touch()
    (lib_dir / "app/support.so").chmod(0o775)

    # An SO file with same group and world permissions
    (lib_dir / "app/support_same_perms.so").touch()
    (lib_dir / "app/support_same_perms.so").chmod(0o744)

    # A SO file with both .so and .so.1.0 forms
    (lib_dir / "app_packages/firstlib/first.so").touch()
    (lib_dir / "app_packages/firstlib/first.so.1.0").touch()

    # An SO file with 664 permissions
    (lib_dir / "app_packages/secondlib/second_a.so").touch()
    (lib_dir / "app_packages/secondlib/second_a.so").chmod(0o664)

    (lib_dir / "app_packages/secondlib/second_b.so").touch()

    return first_app_config



import os
from unittest import mock

import pytest
from cookiecutter import exceptions as cookiecutter_exceptions
from cookiecutter.main import cookiecutter

import briefcase
from briefcase.commands import NewCommand
from briefcase.console import Console, Log
from briefcase.exceptions import (
    BriefcaseCommandError,
    InvalidTemplateRepository,
    TemplateUnsupportedVersion,
)


@pytest.fixture
def new_command(tmp_path):
    return NewCommand(
        base_path=tmp_path / "base",
        data_path=tmp_path / "data",
        logger=Log(),
        console=Console(),
    )


@pytest.mark.parametrize(
    "briefcase_version, expected_branch",
    [
        ("37.42.1", "v37.42.1"),
        ("37.42.2.dev0+gad61a29.d20220919", "v37.42.2"),
        ("37.42.3.dev73+gad61a29.d20220919", "v37.42.3"),
        ("37.42.4a1", "v37.42.4"),
        ("37.42.5b2", "v37.42.5"),
        ("37.42.6rc3", "v37.42.6"),
        ("37.42.7.post1", "v37.42.7"),
    ],
)
def test_new_app(
    monkeypatch,
    new_command,
    tmp_path,
    briefcase_version,
    expected_branch,
):
    """A new app can be created with the default template."""
    monkeypatch.setattr(briefcase, "__version__", briefcase_version)
    app_context = {
        "formal_name": "My Application",
        "class_name": "MyApplication",
        "app_name": "myapplication",
    }
    new_command.build_app_context = mock.MagicMock(return_value=app_context)
    new_command.build_gui_context = mock.MagicMock(
        return_value={
            "app_source": "main()",
            "pyproject_requires": "toga",
        }
    )
    new_command.update_cookiecutter_cache = mock.MagicMock(
        return_value="~/.cookiecutters/briefcase-template"
    )
    new_command.tools.cookiecutter = mock.MagicMock(spec_set=cookiecutter)

    # Create the new app, using the default template.
    new_command.new_app()

    # App context is constructed
    new_command.build_app_context.assert_called_once_with(None)
    new_command.build_gui_context.assert_called_once_with(app_context, None)
    # Template is updated
    new_command.update_cookiecutter_cache.assert_called_once_with(
        template="https://github.com/beeware/briefcase-template",
        branch=expected_branch,
    )
    # Cookiecutter is invoked
    new_command.tools.cookiecutter.assert_called_once_with(
        "~/.cookiecutters/briefcase-template",
        no_input=True,
        output_dir=os.fsdecode(tmp_path / "base"),
        checkout=expected_branch,
        extra_context={
            "formal_name": "My Application",
            "class_name": "MyApplication",
            "app_name": "myapplication",
            # The expected app context
            # should now also contain the
            # default template and branch
            "template_source": "https://github.com/beeware/briefcase-template",
            "template_branch": expected_branch,
            "briefcase_version": briefcase_version,
            "app_source": "main()",
            "pyproject_requires": "toga",
        },
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


def test_new_app_missing_template(monkeypatch, new_command, tmp_path):
    """If a versioned branch doesn't exist, an error is raised."""
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    app_context = {
        "formal_name": "My Application",
        "class_name": "MyApplication",
        "app_name": "myapplication",
    }
    new_command.build_app_context = mock.MagicMock(return_value=app_context)
    new_command.build_gui_context = mock.MagicMock(
        return_value={
            "app_source": "main()",
            "pyproject_requires": "toga",
        }
    )
    new_command.update_cookiecutter_cache = mock.MagicMock(
        return_value="~/.cookiecutters/briefcase-template"
    )
    new_command.tools.cookiecutter = mock.MagicMock()

    # Mock a failed call to cookiecutter when requesting the branch
    new_command.tools.cookiecutter.side_effect = (
        cookiecutter_exceptions.RepositoryCloneFailed
    )

    # Create the new app, using the default template. This raises an error.
    with pytest.raises(TemplateUnsupportedVersion):
        new_command.new_app()

    # App context is constructed
    new_command.build_app_context.assert_called_once_with(None)
    new_command.build_gui_context.assert_called_once_with(app_context, None)

    # The cookiecutter cache is updated once
    new_command.update_cookiecutter_cache.assert_called_once_with(
        template="https://github.com/beeware/briefcase-template",
        branch="v37.42.7",
    )

    # Cookiecutter is invoked twice
    new_command.tools.cookiecutter.assert_called_once_with(
        "~/.cookiecutters/briefcase-template",
        no_input=True,
        output_dir=os.fsdecode(tmp_path / "base"),
        checkout="v37.42.7",
        extra_context={
            "formal_name": "My Application",
            "class_name": "MyApplication",
            "app_name": "myapplication",
            # The expected app context
            # should now also contain the
            # default template and branch
            "template_source": "https://github.com/beeware/briefcase-template",
            "template_branch": "v37.42.7",
            "briefcase_version": "37.42.7",
            "app_source": "main()",
            "pyproject_requires": "toga",
        },
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


@pytest.mark.parametrize(
    "briefcase_version",
    ("37.42.7.dev0+gad61a29.d20220919", "37.42.7.dev73+gad61a29.d20220919"),
)
def test_new_app_dev(monkeypatch, new_command, tmp_path, briefcase_version):
    """In a dev version, template will fall back to the 'main' branch if a versioned
    template doesn't exist."""
    monkeypatch.setattr(briefcase, "__version__", briefcase_version)
    app_context = {
        "formal_name": "My Application",
        "class_name": "MyApplication",
        "app_name": "myapplication",
    }
    new_command.build_app_context = mock.MagicMock(return_value=app_context)
    new_command.build_gui_context = mock.MagicMock(
        return_value={
            "app_source": "main()",
            "pyproject_requires": "toga",
        }
    )
    new_command.update_cookiecutter_cache = mock.MagicMock(
        return_value="~/.cookiecutters/briefcase-template"
    )
    new_command.tools.cookiecutter = mock.MagicMock()

    # There will be two calls to cookiecutter; one on the versioned branch,
    # and one on the `main` branch. The first call will fail because the
    # template doesn't exist yet; the second will succeed.
    new_command.tools.cookiecutter.side_effect = [
        cookiecutter_exceptions.RepositoryCloneFailed,
        None,
    ]

    # Create the new app, using the default template.
    new_command.new_app()

    # App context is constructed
    new_command.build_app_context.assert_called_once_with(None)
    new_command.build_gui_context.assert_called_once_with(app_context, None)
    # Template is updated
    assert new_command.update_cookiecutter_cache.mock_calls == [
        mock.call(
            template="https://github.com/beeware/briefcase-template",
            branch="v37.42.7",
        ),
        mock.call(
            template="https://github.com/beeware/briefcase-template",
            branch="main",
        ),
    ]

    # Cookiecutter is invoked twice
    new_command.tools.cookiecutter.assert_has_calls(
        [
            mock.call(
                "~/.cookiecutters/briefcase-template",
                no_input=True,
                output_dir=os.fsdecode(tmp_path / "base"),
                checkout="v37.42.7",
                extra_context={
                    "formal_name": "My Application",
                    "class_name": "MyApplication",
                    "app_name": "myapplication",
                    # The expected app context should now also contain the default
                    # template and branch
                    "template_source": "https://github.com/beeware/briefcase-template",
                    "template_branch": "v37.42.7",
                    "briefcase_version": briefcase_version,
                    "app_source": "main()",
                    "pyproject_requires": "toga",
                },
                default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
            ),
            mock.call(
                "~/.cookiecutters/briefcase-template",
                no_input=True,
                output_dir=os.fsdecode(tmp_path / "base"),
                checkout="main",
                extra_context={
                    "formal_name": "My Application",
                    "class_name": "MyApplication",
                    "app_name": "myapplication",
                    # The expected app context should now also contain the default
                    # template and branch
                    "template_source": "https://github.com/beeware/briefcase-template",
                    "template_branch": "main",
                    "briefcase_version": briefcase_version,
                    "app_source": "main()",
                    "pyproject_requires": "toga",
                },
                default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
            ),
        ]
    )


def test_new_app_with_template(monkeypatch, new_command, tmp_path):
    """A specific template can be requested."""
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")

    app_context = {
        "formal_name": "My Application",
        "class_name": "MyApplication",
        "app_name": "myapplication",
    }
    new_command.build_app_context = mock.MagicMock(return_value=app_context)
    new_command.build_gui_context = mock.MagicMock(
        return_value={
            "app_source": "main()",
            "pyproject_requires": "toga",
        }
    )
    new_command.update_cookiecutter_cache = mock.MagicMock(
        return_value="https://example.com/other.git"
    )
    new_command.tools.cookiecutter = mock.MagicMock(spec_set=cookiecutter)

    # Create a new app, with a specific template.
    new_command.new_app(template="https://example.com/other.git")

    # App context is constructed
    new_command.build_app_context.assert_called_once_with(None)
    new_command.build_gui_context.assert_called_once_with(app_context, None)
    # Template is updated
    new_command.update_cookiecutter_cache.assert_called_once_with(
        template="https://example.com/other.git",
        branch="v37.42.7",
    )
    # Cookiecutter is invoked
    new_command.tools.cookiecutter.assert_called_once_with(
        "https://example.com/other.git",
        no_input=True,
        output_dir=os.fsdecode(tmp_path / "base"),
        checkout="v37.42.7",
        extra_context={
            "formal_name": "My Application",
            "class_name": "MyApplication",
            "app_name": "myapplication",
            # The expected app context
            # should now also contain the
            # template and branch
            "template_source": "https://example.com/other.git",
            "template_branch": "v37.42.7",
            "briefcase_version": "37.42.7",
            "app_source": "main()",
            "pyproject_requires": "toga",
        },
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


def test_new_app_with_invalid_template(monkeypatch, new_command, tmp_path):
    """If the custom template is invalid, an error is raised."""
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    app_context = {
        "formal_name": "My Application",
        "class_name": "MyApplication",
        "app_name": "myapplication",
    }
    new_command.build_app_context = mock.MagicMock(return_value=app_context)
    new_command.build_gui_context = mock.MagicMock(
        return_value={
            "app_source": "main()",
            "pyproject_requires": "toga",
        }
    )
    new_command.update_cookiecutter_cache = mock.MagicMock(
        return_value="https://example.com/other.git"
    )
    new_command.tools.cookiecutter = mock.MagicMock()

    # Mock an error due to a missing repository
    new_command.tools.cookiecutter.side_effect = (
        cookiecutter_exceptions.RepositoryNotFound
    )

    # Create a new app, with a specific template.
    with pytest.raises(InvalidTemplateRepository):
        new_command.new_app(template="https://example.com/other.git")

    # App context is constructed
    new_command.build_app_context.assert_called_once_with(None)
    new_command.build_gui_context.assert_called_once_with(app_context, None)
    # Template is updated
    new_command.update_cookiecutter_cache.assert_called_once_with(
        template="https://example.com/other.git",
        branch="v37.42.7",
    )
    # Cookiecutter is invoked
    new_command.tools.cookiecutter.assert_called_once_with(
        "https://example.com/other.git",
        no_input=True,
        output_dir=os.fsdecode(tmp_path / "base"),
        checkout="v37.42.7",
        extra_context={
            "formal_name": "My Application",
            "class_name": "MyApplication",
            "app_name": "myapplication",
            # The expected app context
            # should now also contain the
            # template and branch
            "template_source": "https://example.com/other.git",
            "template_branch": "v37.42.7",
            "briefcase_version": "37.42.7",
            "app_source": "main()",
            "pyproject_requires": "toga",
        },
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


def test_new_app_with_invalid_template_branch(monkeypatch, new_command, tmp_path):
    """If the custom template doesn't have a branch for the version, an error is
    raised."""
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    app_context = {
        "formal_name": "My Application",
        "class_name": "MyApplication",
        "app_name": "myapplication",
    }
    new_command.build_app_context = mock.MagicMock(return_value=app_context)
    new_command.build_gui_context = mock.MagicMock(
        return_value={
            "app_source": "main()",
            "pyproject_requires": "toga",
        }
    )
    new_command.update_cookiecutter_cache = mock.MagicMock(
        return_value="https://example.com/other.git"
    )
    new_command.tools.cookiecutter = mock.MagicMock()

    # Mock the error when the branch doesn't exist
    new_command.tools.cookiecutter.side_effect = (
        cookiecutter_exceptions.RepositoryCloneFailed
    )

    # Create a new app, with a specific template; this raises an error
    with pytest.raises(TemplateUnsupportedVersion):
        new_command.new_app(template="https://example.com/other.git")

    # App context is constructed
    new_command.build_app_context.assert_called_once_with(None)
    new_command.build_gui_context.assert_called_once_with(app_context, None)

    # Template is updated
    new_command.update_cookiecutter_cache.assert_called_once_with(
        template="https://example.com/other.git",
        branch="v37.42.7",
    )

    # Cookiecutter is invoked
    new_command.tools.cookiecutter.assert_called_once_with(
        "https://example.com/other.git",
        no_input=True,
        output_dir=os.fsdecode(tmp_path / "base"),
        checkout="v37.42.7",
        extra_context={
            "formal_name": "My Application",
            "class_name": "MyApplication",
            "app_name": "myapplication",
            # The expected app context
            # should now also contain the
            # template and branch
            "template_source": "https://example.com/other.git",
            "template_branch": "v37.42.7",
            "briefcase_version": "37.42.7",
            "app_source": "main()",
            "pyproject_requires": "toga",
        },
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


def test_new_app_with_branch(monkeypatch, new_command, tmp_path):
    """A specific branch can be requested."""
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    app_context = {
        "formal_name": "My Application",
        "class_name": "MyApplication",
        "app_name": "myapplication",
    }
    new_command.build_app_context = mock.MagicMock(return_value=app_context)
    new_command.build_gui_context = mock.MagicMock(
        return_value={
            "app_source": "main()",
            "pyproject_requires": "toga",
        }
    )
    new_command.update_cookiecutter_cache = mock.MagicMock(
        return_value="https://example.com/other.git"
    )
    new_command.tools.cookiecutter = mock.MagicMock(spec_set=cookiecutter)

    # Create a new app, with a specific template branch.
    new_command.new_app(template_branch="experimental")

    # App context is constructed
    new_command.build_app_context.assert_called_once_with(None)
    new_command.build_gui_context.assert_called_once_with(app_context, None)
    # Template is updated
    new_command.update_cookiecutter_cache.assert_called_once_with(
        template="https://github.com/beeware/briefcase-template",
        branch="experimental",
    )
    # Cookiecutter is invoked
    new_command.tools.cookiecutter.assert_called_once_with(
        "https://example.com/other.git",
        no_input=True,
        output_dir=os.fsdecode(tmp_path / "base"),
        checkout="experimental",
        extra_context={
            "formal_name": "My Application",
            "class_name": "MyApplication",
            "app_name": "myapplication",
            # The expected app context
            # should now also contain the
            # template and branch
            "template_source": "https://github.com/beeware/briefcase-template",
            "template_branch": "experimental",
            "briefcase_version": "37.42.7",
            "app_source": "main()",
            "pyproject_requires": "toga",
        },
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


def test_new_app_unused_project_overrides(
    monkeypatch,
    new_command,
    tmp_path,
    capsys,
):
    """The user is informed of unused project configuration overrides."""
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    app_context = {
        "formal_name": "My Application",
        "class_name": "MyApplication",
        "app_name": "myapplication",
    }
    new_command.build_app_context = mock.MagicMock(return_value=app_context)
    new_command.build_gui_context = mock.MagicMock(
        return_value={
            "app_source": "main()",
            "pyproject_requires": "toga",
        }
    )
    new_command.update_cookiecutter_cache = mock.MagicMock(
        return_value="~/.cookiecutters/briefcase-template"
    )
    new_command.tools.cookiecutter = mock.MagicMock(spec_set=cookiecutter)

    # Create the new app, using the default template.
    new_command.new_app(project_overrides={"unused": "override"})

    # App context is constructed
    new_command.build_app_context.assert_called_once_with({"unused": "override"})
    new_command.build_gui_context.assert_called_once_with(
        app_context, {"unused": "override"}
    )
    # Template is updated
    new_command.update_cookiecutter_cache.assert_called_once_with(
        template="https://github.com/beeware/briefcase-template",
        branch="v37.42.7",
    )
    # Cookiecutter is invoked
    new_command.tools.cookiecutter.assert_called_once_with(
        "~/.cookiecutters/briefcase-template",
        no_input=True,
        output_dir=os.fsdecode(tmp_path / "base"),
        checkout="v37.42.7",
        extra_context={
            "formal_name": "My Application",
            "class_name": "MyApplication",
            "app_name": "myapplication",
            # The expected app context
            # should now also contain the
            # default template and branch
            "template_source": "https://github.com/beeware/briefcase-template",
            "template_branch": "v37.42.7",
            "briefcase_version": "37.42.7",
            "app_source": "main()",
            "pyproject_requires": "toga",
        },
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )

    unused_project_override_warning = (
        "WARNING: These project configuration overrides were not used:\n\n"
        "    unused = override"
    )
    assert unused_project_override_warning in capsys.readouterr().out


def test_abort_if_directory_exists(monkeypatch, new_command, tmp_path):
    """If the application name directory exists, the create aborts."""
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")

    # Create a colliding app name.
    (tmp_path / "base/myapplication").mkdir(parents=True)

    app_context = {
        "formal_name": "My Application",
        "class_name": "MyApplication",
        "app_name": "myapplication",
    }
    new_command.build_app_context = mock.MagicMock(return_value=app_context)
    new_command.build_gui_context = mock.MagicMock(
        return_value={
            "app_source": "main()",
            "pyproject_requires": "toga",
        }
    )
    new_command.update_cookiecutter_cache = mock.MagicMock(
        return_value="~/.cookiecutters/briefcase-template"
    )
    new_command.tools.cookiecutter = mock.MagicMock(spec_set=cookiecutter)

    # Create the new app, using the default template.
    # This will raise an error due to the colliding directory
    with pytest.raises(BriefcaseCommandError):
        new_command.new_app()

    # App context is constructed
    new_command.build_app_context.assert_called_once_with(None)
    new_command.build_gui_context.assert_called_once_with(app_context, None)
    # Template won't be updated or unrolled
    # Cookiecutter was *not* invoked
    assert new_command.update_cookiecutter_cache.call_count == 0
    assert new_command.tools.cookiecutter.call_count == 0



from unittest.mock import MagicMock

from ...utils import PartialMatchString


def test_pep621_author(convert_command, monkeypatch):
    mock_input_text = MagicMock()
    mock_input_text.return_value = "Some Name"
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)
    (convert_command.base_path / "pyproject.toml").write_text(
        "[project]\n"
        "authors = [\n"
        '    {name="Firstname Firstauthor", email="mail1@tld.com"},\n'
        '    {email="mail2@tld.com"},\n'
        '    {name="Name Thirdauthor", email="mail3@tld.com"},\n'
        '    {name="Firstname Fourthauthor", email="mail4@tld.com"},\n'
        "]",
        encoding="utf-8",
    )

    convert_command.input_email("Name Thirdauthor", "some.bundle", None)
    mock_input_text.assert_called_once_with(
        intro=PartialMatchString("the selected author name"),
        variable="author's email",
        default="mail3@tld.com",
        validator=convert_command.validate_email,
        override_value=None,
    )


def test_pep621_wrong_author(convert_command, monkeypatch):
    mock_input_text = MagicMock()
    mock_input_text.return_value = "Some Name"
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)
    (convert_command.base_path / "pyproject.toml").write_text(
        "[project]\n"
        "authors = [\n"
        '    {name="Firstname Firstauthor", email="mail1@tld.com"},\n'
        '    {email="mail2@tld.com"},\n'
        '    {name="Name Thirdauthor", email="mail3@tld.com"},\n'
        '    {name="Firstname Fourthauthor", email="mail4@tld.com"},\n'
        "]",
        encoding="utf-8",
    )

    convert_command.input_email("Noname Thirdauthor", "some.bundle", None)
    mock_input_text.assert_called_once_with(
        intro=PartialMatchString("the author name and bundle"),
        variable="author's email",
        default="noname@bundle.some",
        validator=convert_command.validate_email,
        override_value=None,
    )


def test_no_pep621_author(convert_command, monkeypatch):
    mock_input_text = MagicMock()
    mock_input_text.return_value = "Some Name"
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)

    convert_command.input_email("Noname Thirdauthor", "some.bundle", None)
    mock_input_text.assert_called_once_with(
        intro=PartialMatchString("the author name and bundle"),
        variable="author's email",
        default="noname@bundle.some",
        validator=convert_command.validate_email,
        override_value=None,
    )


def test_override(convert_command):
    (convert_command.base_path / "pyproject.toml").write_text(
        "[project]\n"
        "authors = [\n"
        '    {name="Firstname Firstauthor", email="mail1@tld.com"},\n'
        '    {email="mail2@tld.com"},\n'
        '    {name="Name Thirdauthor", email="mail3@tld.com"},\n'
        '    {name="Firstname Fourthauthor", email="mail4@tld.com"},\n'
        "]",
        encoding="utf-8",
    )
    assert (
        convert_command.input_email(
            "Noname Thirdauthor", "some.bundle", "some@email.com"
        )
        == "some@email.com"
    )


def test_prompted_email(convert_command):
    """You can type in the e-mail address."""
    convert_command.input.values = ["my@email.com"]
    assert (
        convert_command.input_email("Some name", "com.some.bundle", None)
        == "my@email.com"
    )



import pytest

from briefcase.exceptions import BriefcaseConfigError

from ...utils import create_file


def test_missing_config(base_command):
    """If the configuration file doesn't exit, raise an error."""
    filename = base_command.base_path / "does_not_exist.toml"
    with pytest.raises(BriefcaseConfigError, match="Configuration file not found"):
        base_command.parse_config(filename, {})


def test_incomplete_global_config(base_command):
    """If the global configuration is missing a required argument, an error is
    raised."""
    # Provide a configuration that is missing `bundle`, a required attribute
    filename = base_command.base_path / "pyproject.toml"
    create_file(
        filename,
        """
        [tool.briefcase]
        version = "1.2.3"
        description = "A sample app"
        license.file = "LICENSE"

        [tool.briefcase.app.my-app]
    """,
    )

    with pytest.raises(
        BriefcaseConfigError,
        match=r"Global configuration is incomplete \(missing 'bundle', 'project_name'\)",
    ):
        base_command.parse_config(filename, {})


def test_incomplete_config(base_command):
    """If the configuration is missing a required argument, an error is raised."""
    # Provide a configuration that is missing `bundle`, a required attribute
    filename = base_command.base_path / "pyproject.toml"
    create_file(
        filename,
        """
        [tool.briefcase]
        project_name = "Sample project"
        version = "1.2.3"
        bundle = "com.example"
        description = "A sample app"
        license.file = "LICENSE"

        [tool.briefcase.app.my-app]
    """,
    )

    with pytest.raises(
        BriefcaseConfigError,
        match=r"Configuration for 'my-app' is incomplete \(missing 'sources'\)",
    ):
        base_command.parse_config(filename, {})


def test_parse_config(base_command):
    """A well-formed configuration file can turned into a set of configurations."""
    filename = base_command.base_path / "pyproject.toml"
    create_file(
        filename,
        """
        [tool.briefcase]
        project_name = "Sample project"
        version = "1.2.3"
        description = "A sample app"
        bundle = "org.beeware"
        mystery = 'default'
        license.file = "LICENSE"

        [tool.briefcase.app.firstapp]
        sources = ['src/firstapp']

        [tool.briefcase.app.secondapp]
        sources = ['src/secondapp']
        extra = 'something'
        mystery = 'sekrits'
    """,
    )

    # Parse the configuration, no overrides
    base_command.parse_config(filename, {})

    # There is a global configuration object
    assert repr(base_command.global_config) == "<Sample project v1.2.3 GlobalConfig>"
    assert base_command.global_config.project_name == "Sample project"
    assert base_command.global_config.bundle == "org.beeware"
    assert base_command.global_config.version == "1.2.3"

    # The first app will have all the base attributes required by an app,
    # defined in the config file.
    assert (
        repr(base_command.apps["firstapp"]) == "<org.beeware.firstapp v1.2.3 AppConfig>"
    )
    assert base_command.apps["firstapp"].project_name == "Sample project"
    assert base_command.apps["firstapp"].app_name == "firstapp"
    assert base_command.apps["firstapp"].bundle == "org.beeware"
    assert base_command.apps["firstapp"].mystery == "default"
    assert not hasattr(base_command.apps["firstapp"], "extra")

    # The second app is much the same, except that it has an override
    # value for `mystery`, and an `extra` value.
    assert (
        repr(base_command.apps["secondapp"])
        == "<org.beeware.secondapp v1.2.3 AppConfig>"
    )
    assert base_command.apps["secondapp"].project_name == "Sample project"
    assert base_command.apps["secondapp"].app_name == "secondapp"
    assert base_command.apps["secondapp"].bundle == "org.beeware"
    assert base_command.apps["secondapp"].mystery == "sekrits"
    assert base_command.apps["secondapp"].extra == "something"


def test_parse_config_with_overrides(base_command):
    """A well-formed configuration file can be augmented by the command line."""
    filename = base_command.base_path / "pyproject.toml"
    create_file(
        filename,
        """
        [tool.briefcase]
        project_name = "Sample project"
        version = "1.2.3"
        description = "A sample app"
        bundle = "org.beeware"
        mystery = 'default'
        license.file = "LICENSE"

        [tool.briefcase.app.firstapp]
        sources = ['src/firstapp']

        [tool.briefcase.app.secondapp]
        sources = ['src/secondapp']
        extra = 'something'
        mystery = 'sekrits'
    """,
    )

    # Parse the configuration, with overrides
    base_command.parse_config(
        filename,
        {
            "version": "2.3.4",
            "custom": "something special",
            "mystery": "overridden",
        },
    )

    # There is a global configuration object. It picks up the override values.
    assert repr(base_command.global_config) == "<Sample project v2.3.4 GlobalConfig>"
    assert base_command.global_config.project_name == "Sample project"
    assert base_command.global_config.bundle == "org.beeware"
    assert base_command.global_config.version == "2.3.4"
    assert base_command.global_config.custom == "something special"
    assert base_command.global_config.mystery == "overridden"

    # The first app will have all the base attributes required by an app,
    # defined in the config file, with the values from the overrides taking precedence.
    assert (
        repr(base_command.apps["firstapp"]) == "<org.beeware.firstapp v2.3.4 AppConfig>"
    )
    assert base_command.apps["firstapp"].project_name == "Sample project"
    assert base_command.apps["firstapp"].app_name == "firstapp"
    assert base_command.apps["firstapp"].bundle == "org.beeware"
    assert base_command.apps["firstapp"].version == "2.3.4"
    assert base_command.apps["firstapp"].custom == "something special"
    assert base_command.apps["firstapp"].mystery == "overridden"
    assert not hasattr(base_command.apps["firstapp"], "extra")

    # The second app is much the same. The value for `mystery`` is overridden by the
    # command line (superseding the app override); but the `extra` app-specific value
    # is preserved.
    assert (
        repr(base_command.apps["secondapp"])
        == "<org.beeware.secondapp v2.3.4 AppConfig>"
    )
    assert base_command.apps["secondapp"].project_name == "Sample project"
    assert base_command.apps["secondapp"].app_name == "secondapp"
    assert base_command.apps["secondapp"].bundle == "org.beeware"
    assert base_command.apps["secondapp"].version == "2.3.4"
    assert base_command.apps["secondapp"].custom == "something special"
    assert base_command.apps["secondapp"].mystery == "overridden"
    assert base_command.apps["secondapp"].extra == "something"


def test_parse_config_with_invalid_override(base_command):
    """If an override value doesn't pass validation, an exception is raised."""
    filename = base_command.base_path / "pyproject.toml"
    create_file(
        filename,
        """
        [tool.briefcase]
        project_name = "Sample project"
        version = "1.2.3"
        description = "A sample app"
        bundle = "org.beeware"
        mystery = 'default'
        license.file = "LICENSE"

        [tool.briefcase.app.firstapp]
        sources = ['src/firstapp']
    """,
    )

    # Parse the configuration, with overrides
    with pytest.raises(
        BriefcaseConfigError,
        match=r"Version numbers must be PEP440 compliant",
    ):
        base_command.parse_config(
            filename,
            {
                "version": "not-a-version-number",
            },
        )



import pytest

from briefcase.console import select_option
from tests.utils import DummyConsole


def test_select_option():
    # Return '3' when prompted
    console = DummyConsole("3")

    options = {
        "first": "The first option",
        "second": "The second option",
        "third": "The third option",
        "fourth": "The fourth option",
    }
    result = select_option(options, input=console)

    # Input is requested once
    assert console.prompts == ["> "]

    # Alphabetically, option 3 will be "the second option"
    assert result == "second"


def test_select_option_list():
    """If select_option is given a list of tuples, they're presented as provided."""
    # Return '3' when prompted
    console = DummyConsole("3")

    options = [
        ("first", "The first option"),
        ("second", "The second option"),
        ("third", "The third option"),
        ("fourth", "The fourth option"),
    ]
    result = select_option(options, input=console)

    # Input is requested once
    assert console.prompts == ["> "]

    # The third option is the third option :-)
    assert result == "third"


def test_select_option_bad_input():
    # In order, return:
    #     blank
    #     'asdf'
    #     '10'
    #     '3'
    console = DummyConsole("", "asdf", "10", "3")

    options = {
        "first": "The first option",
        "second": "The second option",
        "third": "The third option",
        "fourth": "The fourth option",
    }
    result = select_option(options, input=console)

    # Input is requested five times; first four cause errors.
    assert console.prompts == ["> "] * 4

    # Alphabetically, option 3 will be "the second option"
    assert result == "second"


@pytest.mark.parametrize("default, expected_choice", [("1", "first"), ("3", "third")])
def test_select_option_default(default, expected_choice):
    """If select_option has a default, it is returned for no input."""
    # Return an empty response when prompted as though the user press entered
    console = DummyConsole("")

    options = [
        ("first", "The first option"),
        ("second", "The second option"),
        ("third", "The third option"),
        ("fourth", "The fourth option"),
    ]
    result = select_option(options, input=console, default=default)

    # Input is requested once
    assert console.prompts == ["> "]

    # The default option is returned
    assert result == expected_choice



import subprocess
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import BriefcaseCommandError


@pytest.mark.parametrize(
    "ldd_output, version",
    [
        # ubuntu:focal
        (
            [
                "ldd (Ubuntu GLIBC 2.31-0ubuntu9.9) 2.31",
                "Copyright (C) 2020 Free Software Foundation, Inc.",
                "...",
            ],
            "2.31",
        ),
        # fedora:37
        (
            [
                "ldd (GNU libc) 2.36",
                "Copyright (C) 2020 Free Software Foundation, Inc.",
                "...",
            ],
            "2.36",
        ),
        # arch:rolling (20230301)
        (
            [
                "ldd (GNU libc) 2.37",
                "Copyright (C) 2023 Free Software Foundation, Inc.",
                "...",
            ],
            "2.37",
        ),
    ],
)
def test_target_glibc_version_docker(
    create_command, first_app_config, ldd_output, version
):
    "Test that the glibc version in a docker container can be determined"
    # Mock an app being built on docker
    create_command.target_image = "somevendor:surprising"
    first_app_config.target_image = "somevendor:surprising"

    # Mock a verified Docker, and the output of ldd.
    create_command.tools.docker = MagicMock()
    create_command.tools.docker.check_output.return_value = "\n".join(ldd_output)

    # The glibc version was returned
    assert create_command.target_glibc_version(first_app_config) == version

    # Docker was consulted for the glibc version
    create_command.tools.docker.check_output.assert_called_once_with(
        ["ldd", "--version"],
        image_tag="somevendor:surprising",
    )


def test_target_glibc_version_docker_no_ldd(create_command, first_app_config):
    "If the Docker container can't run ldd, an error is raised"
    # Mock an app being built on docker
    create_command.target_image = "somevendor:surprising"
    first_app_config.target_image = "somevendor:surprising"

    # Mock a verified Docker, and an error from invoking ldd.
    create_command.tools.docker = MagicMock()
    create_command.tools.docker.check_output.side_effect = (
        subprocess.CalledProcessError(cmd=["docker ..."], returncode=-1)
    )

    # An error is raised when getting the glibc version
    with pytest.raises(
        BriefcaseCommandError, match=r"Unable to determine glibc dependency version."
    ):
        create_command.target_glibc_version(first_app_config)

    # Docker was consulted for the glibc version
    create_command.tools.docker.check_output.assert_called_once_with(
        ["ldd", "--version"],
        image_tag="somevendor:surprising",
    )


def test_target_glibc_version_docker_bad_ldd_output(create_command, first_app_config):
    "If ldd returns unexpected content, an error is raised"
    # Mock an app being built on docker
    create_command.target_image = "somevendor:surprising"
    first_app_config.target_image = "somevendor:surprising"

    # Mock a verified Docker, and unexpected ldd output
    create_command.tools.docker = MagicMock()
    create_command.tools.docker.check_output.return_value = (
        "I don't know what this is, but it isn't ldd output."
    )

    # An error is raised when getting the glibc version
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to parse glibc dependency version from version string.",
    ):
        create_command.target_glibc_version(first_app_config)

    # Docker was consulted for the glibc version
    create_command.tools.docker.check_output.assert_called_once_with(
        ["ldd", "--version"],
        image_tag="somevendor:surprising",
    )


def test_target_glibc_version_nodocker(create_command, first_app_config):
    "Test that the glibc version of the local system can be returned"
    # Mock a non-docker setup
    create_command.target_image = None
    create_command.tools.os.confstr = MagicMock(return_value="glibc 2.42")

    # The glibc version was returned
    assert create_command.target_glibc_version(first_app_config) == "2.42"

    # The OS module was consulted for the glibc version
    create_command.tools.os.confstr.assert_called_once_with("CS_GNU_LIBC_VERSION")



import subprocess
from unittest import mock


def test_clean_termination(mock_sub, capsys):
    """A popen process can be cleanly terminated."""
    process = mock.MagicMock(spec_set=subprocess.Popen)

    mock_sub.cleanup("testing process", process)

    process.terminate.assert_called_once()
    process.wait.assert_called_once_with(timeout=3)
    process.kill.assert_not_called()

    # No log messages for a clean exit
    assert capsys.readouterr().out == ""


def test_dirty_termination(mock_sub, capsys):
    """If terminate doesn't stop the process, it will be forcibly killed."""
    process = mock.MagicMock(spec_set=subprocess.Popen)
    process.wait.side_effect = subprocess.TimeoutExpired(cmd="ls", timeout=3)

    mock_sub.cleanup("testing process", process)

    process.terminate.assert_called_once()
    process.wait.assert_called_once_with(timeout=3)
    process.kill.assert_called_once_with()

    # Log contains a contextual message.
    assert capsys.readouterr().out == "Forcibly killing testing process...\n"



import subprocess

import pytest

from briefcase.exceptions import BriefcaseCommandError

from ...utils import create_file, create_installed_package, file_content


def test_thin_app_packages(dummy_command, tmp_path):
    "An app packages folder can be thinned"
    app_packages = tmp_path / "app_pacakges.gothic"

    # Install a package into the app_packages folder that contains source and binaries.
    create_installed_package(
        app_packages,
        "pkg",
        "2.3.4",
        tag="macOS_11_0_gothic",
        extra_content=[
            ("pkg/other.py", "# other python"),
            ("pkg/sub1/module1.dylib", "dylib-fat"),
            ("pkg/sub1/module2.so", "dylib-fat"),
            ("pkg/sub2/module3.dylib", "dylib-fat"),
        ],
    )

    # All dylibs have 2 architectures
    dummy_command.tools.subprocess.check_output.return_value = (
        "Architectures in the fat file: path/to/file.dylib are: modern gothic\n"
    )

    # Mock the effect of calling lipo -thin
    def thin_dylib(*args, **kwargs):
        create_file(args[0][args[0].index("-output") + 1], "dylib-thin")

    dummy_command.tools.subprocess.run.side_effect = thin_dylib

    # Thin the app_packages folder to gothic dylibs
    dummy_command.thin_app_packages(app_packages, arch="gothic")

    # All libraries have been thinned
    assert file_content(app_packages / "pkg/sub1/module1.dylib") == "dylib-thin"
    assert file_content(app_packages / "pkg/sub1/module2.so") == "dylib-thin"
    assert file_content(app_packages / "pkg/sub2/module3.dylib") == "dylib-thin"


def test_thin_app_packages_problem(dummy_command, tmp_path):
    "If one of the libraries can't be thinned, an error is raised"
    app_packages = tmp_path / "app_pacakges.gothic"

    # Install a package into the app_packages folder that contains source and binaries.
    create_installed_package(
        app_packages,
        "pkg",
        "2.3.4",
        tag="macOS_11_0_gothic",
        extra_content=[
            ("pkg/other.py", "# other python"),
            ("pkg/sub1/module1.dylib", "dylib-fat"),
            ("pkg/sub1/module2.so", "dylib-fat"),
            ("pkg/sub2/module3.dylib", "dylib-fat"),
        ],
    )

    # All dylibs have 2 architectures
    dummy_command.tools.subprocess.check_output.return_value = (
        "Architectures in the fat file: path/to/file.dylib are: modern gothic\n"
    )

    # Mock the effect of calling lipo -thin. Calling on a .so file raises an error.
    def thin_dylib(*args, **kwargs):
        if str(args[0][-1]).endswith(".so"):
            raise subprocess.CalledProcessError(cmd="lipo -thin", returncode=-1)
        create_file(args[0][args[0].index("-output") + 1], "dylib-thin")

    dummy_command.tools.subprocess.run.side_effect = thin_dylib

    # Thin the app_packages folder to gothic dylibs. This raises an error:
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to create thin binary from .*module2\.so",
    ):
        dummy_command.thin_app_packages(app_packages, arch="gothic")


def test_thin_no_dylibs(dummy_command, tmp_path):
    "If there are no dylibs, thinning is a no-op."
    app_packages = tmp_path / "app_pacakges.gothic"

    # Install a package into the app_packages folder that only contains source.
    create_installed_package(
        app_packages,
        "pkg",
        "2.3.4",
        tag="macOS_11_0_gothic",
        extra_content=[
            ("pkg/other.py", "# other python"),
        ],
    )

    # Thin the app packages folder
    dummy_command.thin_app_packages(app_packages, arch="gothic")

    # lipo was not called.
    dummy_command.tools.subprocess.check_output.assert_not_called()
    dummy_command.tools.subprocess.run.assert_not_called()



from io import BytesIO
from unittest.mock import Mock, call

import pytest

from briefcase.config import parse_config
from briefcase.exceptions import BriefcaseConfigError


def test_invalid_toml():
    """If the config file isn't TOML, raise an error."""
    config_file = BytesIO(b"this is not toml!")

    with pytest.raises(BriefcaseConfigError, match="Invalid pyproject.toml"):
        parse_config(
            config_file,
            platform="macOS",
            output_format="Xcode",
            logger=Mock(),
        )


def test_no_briefcase_section():
    """If the config file doesn't contain a briefcase tool section, raise an error."""
    config_file = BytesIO(
        b"""
        [tool.section]
        name="value"
        number=42
        """
    )

    with pytest.raises(BriefcaseConfigError, match="No tool.briefcase section"):
        parse_config(
            config_file,
            platform="macOS",
            output_format="Xcode",
            logger=Mock(),
        )


def test_no_apps():
    """If the config file doesn't contain at least one briefcase app, raise an error."""
    config_file = BytesIO(
        b"""
        [tool.briefcase]
        name="value"
        number=42
        """
    )

    with pytest.raises(BriefcaseConfigError, match="No Briefcase apps defined"):
        parse_config(
            config_file,
            platform="macOS",
            output_format="Xcode",
            logger=Mock(),
        )


def test_single_minimal_app():
    """A single app can be defined, but can exist without any app attributes."""
    config_file = BytesIO(
        b"""
        [tool.briefcase]
        value = 42
        license.file = "LICENSE"

        [tool.briefcase.app.my_app]
        """
    )

    global_options, apps = parse_config(
        config_file,
        platform="macOS",
        output_format="Xcode",
        logger=Mock(),
    )

    # There's a single global option
    assert global_options == {"value": 42, "license": {"file": "LICENSE"}}

    # The app gets the name from its header line.
    # It inherits the value from the base definition.
    assert apps == {
        "my_app": {
            "app_name": "my_app",
            "value": 42,
            "license": {"file": "LICENSE"},
        }
    }


def test_multiple_minimal_apps():
    """The configuration can contain multiple apps without an explicit tool header."""
    config_file = BytesIO(
        b"""
        [tool.briefcase.app.first]
        number=37
        license.file = "LICENSE"

        [tool.briefcase.app.second]
        app_name="my_app"
        number=42
        license.file = "LICENSE.txt"
        """
    )

    global_options, apps = parse_config(
        config_file,
        platform="macOS",
        output_format="Xcode",
        logger=Mock(),
    )

    # There are no global options
    assert global_options == {}

    # The apps get their name from the header lines.
    # The second tool overrides its app name
    assert apps == {
        "first": {"app_name": "first", "number": 37, "license": {"file": "LICENSE"}},
        "second": {
            "app_name": "my_app",
            "number": 42,
            "license": {"file": "LICENSE.txt"},
        },
    }


def test_platform_override():
    """An app can define platform settings that override base settings."""
    config_file = BytesIO(
        b"""
        [tool.briefcase]
        value = 0
        basevalue = "the base"
        license.file = "LICENSE"

        [tool.briefcase.app.my_app]
        value = 1
        appvalue = "the app"

        [tool.briefcase.app.my_app.macOS]
        value = 2
        platformvalue = "macos platform"

        [tool.briefcase.app.my_app.linux]
        value = 3
        platformvalue = "linux platform"

        [tool.briefcase.app.other_app.macOS]
        value = 4
        platformvalue = "other macos platform"
        """
    )

    global_options, apps = parse_config(
        config_file,
        platform="macOS",
        output_format="Xcode",
        logger=Mock(),
    )

    # The global options are exactly as specified
    assert global_options == {
        "value": 0,
        "basevalue": "the base",
        "license": {"file": "LICENSE"},
    }

    # Since a macOS app has been requested, the macOS platform values
    # take priority. Linux configuration values are dropped.
    # The second app doesn't provide an explicit app-level config, but
    # the app exists because the platform exists.
    # Platforms should be processed in sorted order, which means that linux
    # will be processed before macOS.
    assert apps == {
        "my_app": {
            "app_name": "my_app",
            "value": 2,
            "basevalue": "the base",
            "appvalue": "the app",
            "platformvalue": "macos platform",
            "license": {"file": "LICENSE"},
        },
        "other_app": {
            "app_name": "other_app",
            "value": 4,
            "basevalue": "the base",
            "platformvalue": "other macos platform",
            "license": {"file": "LICENSE"},
        },
    }


def test_platform_override_ordering():
    """The order of platform processing doesn't affect output."""
    config_file = BytesIO(
        b"""
        [tool.briefcase]
        value = 0
        basevalue = "the base"
        license.file = "LICENSE"

        [tool.briefcase.app.my_app]
        value = 1
        appvalue = "the app"

        [tool.briefcase.app.my_app.macOS]
        value = 2
        platformvalue = "macos platform"

        [tool.briefcase.app.my_app.windows]
        value = 3
        platformvalue = "windows platform"

        [tool.briefcase.app.other_app.macOS]
        value = 4
        platformvalue = "other macos platform"
        """
    )

    global_options, apps = parse_config(
        config_file,
        platform="macOS",
        output_format="Xcode",
        logger=Mock(),
    )

    # The global options are exactly as specified
    assert global_options == {
        "value": 0,
        "basevalue": "the base",
        "license": {"file": "LICENSE"},
    }

    # Since a macOS app has been requested, the macOS platform values
    # take priority. Linux configuration values are dropped.
    # The second app doesn't provide an explicit app-level config, but
    # the app exists because the platform exists.
    # Platforms should be processed in order, which means that windows
    # will be processed after macOS.
    assert apps == {
        "my_app": {
            "app_name": "my_app",
            "value": 2,
            "basevalue": "the base",
            "appvalue": "the app",
            "platformvalue": "macos platform",
            "license": {"file": "LICENSE"},
        },
        "other_app": {
            "app_name": "other_app",
            "value": 4,
            "basevalue": "the base",
            "platformvalue": "other macos platform",
            "license": {"file": "LICENSE"},
        },
    }


def test_format_override():
    """An app can define format settings that override base and platform settings."""
    config_file = BytesIO(
        b"""
        [tool.briefcase]
        value = 0
        basevalue = "the base"
        license.file = "LICENSE"

        [tool.briefcase.app.my_app]
        value = 1
        appvalue = "the app"

        [tool.briefcase.app.my_app.macOS]
        value = 2
        platformvalue = "macos platform"

        [tool.briefcase.app.my_app.macOS.app]
        value = 21
        formatvalue = "app format"

        [tool.briefcase.app.my_app.macOS.Xcode]
        value = 22
        formatvalue = "xcode format"

        [tool.briefcase.app.my_app.linux]
        value = 3
        platformvalue = "linux platform"

        [tool.briefcase.app.my_app.linux.snap]
        value = 31
        formatvalue = "snap format"

        [tool.briefcase.app.my_app.linux.appimage]
        value = 32
        formatvalue = "appimage format"

        [tool.briefcase.app.other_app.macOS.app]
        value = 41
        formatvalue = "other macos app format"
        """
    )

    global_options, apps = parse_config(
        config_file,
        platform="macOS",
        output_format="app",
        logger=Mock(),
    )

    # The global options are exactly as specified
    assert global_options == {
        "value": 0,
        "basevalue": "the base",
        "license": {"file": "LICENSE"},
    }

    # Since a macOS app has been requested, the macOS app format values
    # take priority. Linux configuration values are dropped.
    # The second app doesn't provide an explicit app-level config, but
    # the app exists because the platform exists.
    # Formats should be processed in order, which means that app
    # will be processed before dmg.
    assert apps == {
        "my_app": {
            "app_name": "my_app",
            "value": 21,
            "basevalue": "the base",
            "appvalue": "the app",
            "platformvalue": "macos platform",
            "formatvalue": "app format",
            "license": {"file": "LICENSE"},
        },
        "other_app": {
            "app_name": "other_app",
            "value": 41,
            "basevalue": "the base",
            "formatvalue": "other macos app format",
            "license": {"file": "LICENSE"},
        },
    }


def test_format_override_ordering():
    """The order of format processing doesn't affect output."""
    config_file = BytesIO(
        b"""
        [tool.briefcase]
        value = 0
        basevalue = "the base"
        license.file = "LICENSE"

        [tool.briefcase.app.my_app]
        value = 1
        appvalue = "the app"

        [tool.briefcase.app.my_app.macOS]
        value = 2
        platformvalue = "macos platform"

        [tool.briefcase.app.my_app.macOS.Xcode]
        value = 21
        formatvalue = "app format"

        [tool.briefcase.app.my_app.macOS.app]
        value = 22
        formatvalue = "app format"

        [tool.briefcase.app.my_app.linux]
        value = 3
        platformvalue = "linux platform"

        [tool.briefcase.app.my_app.linux.snap]
        value = 31
        formatvalue = "snap format"

        [tool.briefcase.app.my_app.linux.appimage]
        value = 32
        formatvalue = "appimage format"

        [tool.briefcase.app.other_app.macOS.Xcode]
        value = 41
        formatvalue = "other macos app format"
        """
    )

    global_options, apps = parse_config(
        config_file,
        platform="macOS",
        output_format="app",
        logger=Mock(),
    )

    # The global options are exactly as specified
    assert global_options == {
        "value": 0,
        "basevalue": "the base",
        "license": {"file": "LICENSE"},
    }

    # Since a macOS dmg has been requested, the macOS dmg format values
    # take priority. Linux configuration values are dropped.
    # The second app doesn't provide an explicit app-level config, but
    # the app exists because the platform exists.
    # Formats should be processed in order, which means that dmg
    # will be processed after app.
    assert apps == {
        "my_app": {
            "app_name": "my_app",
            "value": 22,
            "basevalue": "the base",
            "appvalue": "the app",
            "platformvalue": "macos platform",
            "formatvalue": "app format",
            "license": {"file": "LICENSE"},
        },
        "other_app": {
            "app_name": "other_app",
            "value": 0,
            "basevalue": "the base",
            "license": {"file": "LICENSE"},
        },
    }


def test_requires():
    """Requirements can be specified."""
    config_file = BytesIO(
        b"""
        [tool.briefcase]
        value = 0
        requires = ["base value"]
        license.file = "LICENSE"

        [tool.briefcase.app.my_app]
        requires = ["my_app value"]

        [tool.briefcase.app.my_app.macOS]
        requires = ["macos value"]

        [tool.briefcase.app.my_app.macOS.Xcode]
        requires = ["xcode value"]

        [tool.briefcase.app.my_app.macOS.app]
        requires = ["app value"]

        [tool.briefcase.app.my_app.linux]
        requires = ["linux value"]

        [tool.briefcase.app.my_app.linux.appimage]
        requires = ["appimage value"]

        [tool.briefcase.app.other_app]
        """
    )

    # Request a macOS app
    global_options, apps = parse_config(
        config_file,
        platform="macOS",
        output_format="app",
        logger=Mock(),
    )

    # The global options are exactly as specified
    assert global_options == {
        "value": 0,
        "requires": ["base value"],
        "license": {"file": "LICENSE"},
    }

    # The macOS my_app app specifies a full inherited chain.
    # The other_app app doesn't specify any options.
    assert apps == {
        "my_app": {
            "app_name": "my_app",
            "requires": [
                "base value",
                "my_app value",
                "macos value",
                "app value",
            ],
            "value": 0,
            "license": {"file": "LICENSE"},
        },
        "other_app": {
            "app_name": "other_app",
            "requires": [
                "base value",
            ],
            "value": 0,
            "license": {"file": "LICENSE"},
        },
    }

    # Request a macOS xcode project
    config_file.seek(0)
    global_options, apps = parse_config(
        config_file,
        platform="macOS",
        output_format="Xcode",
        logger=Mock(),
    )

    # The global options are exactly as specified
    assert global_options == {
        "value": 0,
        "requires": ["base value"],
        "license": {"file": "LICENSE"},
    }

    # The macOS my_app dmg specifies a full inherited chain.
    # The other_app dmg doesn't specify any options.
    assert apps == {
        "my_app": {
            "app_name": "my_app",
            "requires": [
                "base value",
                "my_app value",
                "macos value",
                "xcode value",
            ],
            "value": 0,
            "license": {"file": "LICENSE"},
        },
        "other_app": {
            "app_name": "other_app",
            "requires": [
                "base value",
            ],
            "value": 0,
            "license": {"file": "LICENSE"},
        },
    }

    config_file.seek(0)
    global_options, apps = parse_config(
        config_file,
        platform="linux",
        output_format="appimage",
        logger=Mock(),
    )

    # The global options are exactly as specified
    assert global_options == {
        "value": 0,
        "requires": ["base value"],
        "license": {"file": "LICENSE"},
    }

    # The linux my_app appimage overrides the *base* value, but extends for linux.
    assert apps == {
        "my_app": {
            "app_name": "my_app",
            "requires": [
                "base value",
                "my_app value",
                "linux value",
                "appimage value",
            ],
            "value": 0,
            "license": {"file": "LICENSE"},
        },
        "other_app": {
            "app_name": "other_app",
            "requires": [
                "base value",
            ],
            "value": 0,
            "license": {"file": "LICENSE"},
        },
    }


def test_document_types():
    """Document types can be specified."""
    config_file = BytesIO(
        b"""
        [tool.briefcase]
        value = 0
        license.file = "LICENSE"

        [tool.briefcase.app.my_app]

        [tool.briefcase.app.my_app.macOS]

        [tool.briefcase.app.my_app.macOS.document_type.document]
        extension = "doc"
        description = "A document"

        [tool.briefcase.app.my_app.macOS.document_type.image]
        extension = "img"
        description = "An image"

        [tool.briefcase.app.other_app]

        """
    )

    # Request a macOS app
    global_options, apps = parse_config(
        config_file,
        platform="macOS",
        output_format="Xcode",
        logger=Mock(),
    )

    # The macOS my_app app specifies a full inherited chain.
    # The other_app app doesn't specify any options.
    assert apps == {
        "my_app": {
            "app_name": "my_app",
            "document_type": {
                "document": {
                    "extension": "doc",
                    "description": "A document",
                },
                "image": {
                    "extension": "img",
                    "description": "An image",
                },
            },
            "value": 0,
            "license": {"file": "LICENSE"},
        },
        "other_app": {
            "app_name": "other_app",
            "value": 0,
            "license": {"file": "LICENSE"},
        },
    }


def test_pep621_defaults():
    config_file = BytesIO(
        b"""
        [project]
        name = "awesome"
        version = "1.2.3"
        authors = [{name = "Kim Park", email = "kim@example.com"}]
        dependencies = ["numpy"]
        description = "awesome project"

        [project.urls]
        Homepage = "https://example.com/awesome"

        [project.optional-dependencies]
        test = ["pytest"]

        [project.license]
        text = "You can use it while standing on one foot"

        [tool.briefcase]
        project_name = "Awesome app"
        bundle = "com.example"

        [tool.briefcase.app.awesome]
        formal_name = "Awesome Application"
        long_description = "The application is very awesome"
        sources = [
            "src",
        ]
        test_sources = [
            "tests",
        ]

        [tool.briefcase.app.awesome.macOS]
        requires = [
            "toga-cocoa~=0.3.1",
            "std-nslog~=1.0.0"
        ]
        """
    )

    global_options, apps = parse_config(
        config_file,
        platform="macOS",
        output_format="app",
        logger=Mock(),
    )

    awesome = apps["awesome"]
    assert awesome == {
        "project_name": "Awesome app",
        "bundle": "com.example",
        "version": "1.2.3",
        "license": {"text": "You can use it while standing on one foot"},
        "author": "Kim Park",
        "author_email": "kim@example.com",
        "url": "https://example.com/awesome",
        "description": "awesome project",
        "requires": ["numpy", "toga-cocoa~=0.3.1", "std-nslog~=1.0.0"],
        "test_requires": ["pytest"],
        "app_name": "awesome",
        "sources": ["src"],
        "test_sources": ["tests"],
        "formal_name": "Awesome Application",
        "long_description": "The application is very awesome",
    }


def test_license_is_string_project():
    """The project definition contains a string definition for 'license'."""
    config_file = BytesIO(
        b"""
        [tool.briefcase]
        value = 0
        license = "Some license"

        [tool.briefcase.app.my_app]
        appvalue = "the app"
        """
    )

    logger = Mock()
    global_options, apps = parse_config(
        config_file, platform="macOS", output_format="app", logger=logger
    )

    assert global_options == {
        "value": 0,
        "license": {"file": "LICENSE"},
    }
    assert apps["my_app"] == {
        "app_name": "my_app",
        "value": 0,
        "appvalue": "the app",
        "license": {"file": "LICENSE"},
    }
    logger.warning.assert_called_once_with(
        """
*************************************************************************
** WARNING: License Definition for the Project is Deprecated           **
*************************************************************************

    Briefcase now uses PEP 621 format for license definitions.

    Previously, the name of the license was assigned to the 'license'
    field in pyproject.toml. For PEP 621, the name of the license is
    assigned to 'license.text' or the name of the file containing the
    license is assigned to 'license.file'.

    The current configuration for the Project has a 'license' field
    that is specified as a string:

        license = "Some license"

    To use the PEP 621 format (and to remove this warning), specify that
    the LICENSE file contains the license for the Project:

        license.file = "LICENSE"

*************************************************************************
"""
    )


def test_license_is_string_project_and_app():
    """The project and app definition contain a string definition for 'license'."""
    config_file = BytesIO(
        b"""
        [tool.briefcase]
        value = 0
        license = "Some license"

        [tool.briefcase.app.my_app]
        appvalue = "the app"
        license = "Another license"
        """
    )

    logger = Mock()
    global_options, apps = parse_config(
        config_file, platform="macOS", output_format="app", logger=logger
    )

    assert global_options == {
        "value": 0,
        "license": {"file": "LICENSE"},
    }
    assert apps["my_app"] == {
        "app_name": "my_app",
        "value": 0,
        "appvalue": "the app",
        "license": {"file": "LICENSE"},
    }
    logger.warning.assert_has_calls(
        [
            call(
                """
*************************************************************************
** WARNING: License Definition for the Project is Deprecated           **
*************************************************************************

    Briefcase now uses PEP 621 format for license definitions.

    Previously, the name of the license was assigned to the 'license'
    field in pyproject.toml. For PEP 621, the name of the license is
    assigned to 'license.text' or the name of the file containing the
    license is assigned to 'license.file'.

    The current configuration for the Project has a 'license' field
    that is specified as a string:

        license = "Some license"

    To use the PEP 621 format (and to remove this warning), specify that
    the LICENSE file contains the license for the Project:

        license.file = "LICENSE"

*************************************************************************
"""
            ),
            call(
                """
*************************************************************************
** WARNING: License Definition for 'my_app' is Deprecated              **
*************************************************************************

    Briefcase now uses PEP 621 format for license definitions.

    Previously, the name of the license was assigned to the 'license'
    field in pyproject.toml. For PEP 621, the name of the license is
    assigned to 'license.text' or the name of the file containing the
    license is assigned to 'license.file'.

    The current configuration for 'my_app' has a 'license' field
    that is specified as a string:

        license = "Another license"

    To use the PEP 621 format (and to remove this warning), specify that
    the LICENSE file contains the license for 'my_app':

        license.file = "LICENSE"

*************************************************************************
"""
            ),
        ]
    )



import subprocess
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import BriefcaseCommandError, InvalidDeviceError


def test_force_stop_app(adb, capsys):
    """Invoking `force_stop_app()` calls `run()` with the appropriate parameters."""
    # Mock out the run command on an adb instance
    adb.run = MagicMock(return_value="example normal adb output")

    # Invoke force_stop_app
    adb.force_stop_app("com.example.sample.package")

    # Validate call parameters.
    adb.run.assert_called_once_with(
        "shell", "am", "force-stop", "com.example.sample.package"
    )

    # Validate that the normal output of the command was not printed (since there
    # was no error).
    assert "normal adb output" not in capsys.readouterr()


def test_force_top_fail(adb, capsys):
    """If `force_stop_app()` fails, an error is raised."""
    # Mock out the run command on an adb instance
    adb.run = MagicMock(
        side_effect=subprocess.CalledProcessError(returncode=69, cmd="force-stop")
    )

    # Invoke force_stop_app
    with pytest.raises(BriefcaseCommandError):
        adb.force_stop_app("com.example.sample.package")

    # Validate call parameters.
    adb.run.assert_called_once_with(
        "shell", "am", "force-stop", "com.example.sample.package"
    )


def test_invalid_device(adb, capsys):
    """Invoking `force_stop_app()` on an invalid device raises an error."""
    # Mock out the run command on an adb instance
    adb.run = MagicMock(side_effect=InvalidDeviceError("device", "exampleDevice"))

    # Invoke force_stop_app
    with pytest.raises(InvalidDeviceError):
        adb.force_stop_app("com.example.sample.package")

    # Validate call parameters.
    adb.run.assert_called_once_with(
        "shell", "am", "force-stop", "com.example.sample.package"
    )



from unittest import mock

import pytest

from briefcase.commands.run import LogFilter
from briefcase.exceptions import BriefcaseCommandError, BriefcaseTestSuiteFailure


def test_run_app(run_command, first_app):
    """An app can have its logs streamed."""
    popen = mock.MagicMock()
    popen.poll = mock.MagicMock(return_value=0)
    clean_filter = mock.MagicMock()

    # Invoke the stop func as part of streaming. This is to satisfy coverage.
    def mock_stream(label, popen_process, stop_func, filter_func):
        stop_func()

    run_command.tools.subprocess.stream_output = mock.MagicMock(side_effect=mock_stream)

    # Stream the app logs
    run_command._stream_app_logs(
        first_app,
        popen=popen,
        test_mode=False,
        clean_filter=clean_filter,
        clean_output=False,
    )

    # The log was streamed
    run_command.tools.subprocess.stream_output.assert_called_once_with(
        label="first",
        popen_process=popen,
        filter_func=mock.ANY,
        stop_func=mock.ANY,
    )

    # The filter function has the properties we'd expect
    filter_func = run_command.tools.subprocess.stream_output.mock_calls[0].kwargs[
        "filter_func"
    ]
    assert not filter_func.clean_output
    assert filter_func.clean_filter == clean_filter
    assert filter_func.exit_filter.regex.pattern == LogFilter.DEFAULT_EXIT_REGEX


def test_run_app_custom_stop_func(run_command, first_app):
    """An app with a custom stop function can have its logs streamed."""
    popen = mock.MagicMock()
    popen.poll = mock.MagicMock(return_value=0)
    clean_filter = mock.MagicMock()
    stop_func = mock.MagicMock()
    run_command.tools.subprocess.stream_output = mock.MagicMock()

    # Stream the app logs
    run_command._stream_app_logs(
        first_app,
        popen=popen,
        test_mode=False,
        clean_filter=clean_filter,
        clean_output=False,
        stop_func=stop_func,
    )

    # The log was streamed
    run_command.tools.subprocess.stream_output.assert_called_once_with(
        label="first",
        popen_process=popen,
        filter_func=mock.ANY,
        stop_func=stop_func,
    )

    # The filter function has the properties we'd expect
    filter_func = run_command.tools.subprocess.stream_output.mock_calls[0].kwargs[
        "filter_func"
    ]
    assert not filter_func.clean_output
    assert filter_func.clean_filter == clean_filter
    assert filter_func.exit_filter.regex.pattern == LogFilter.DEFAULT_EXIT_REGEX


def test_test_mode_success(run_command, first_app):
    """An app can be streamed in test mode."""
    popen = mock.MagicMock()
    popen.returncode = 0
    clean_filter = mock.MagicMock()
    stop_func = mock.MagicMock()
    run_command.tools.subprocess.stream_output = mock.MagicMock()

    # Mock effect of streaming the app resulting in a test suite success
    def mock_stream_output(label, popen_process, filter_func, **kwargs):
        filter_func.returncode = 0

    run_command.tools.subprocess.stream_output.side_effect = mock_stream_output

    # Stream the app logs
    run_command._stream_app_logs(
        first_app,
        popen=popen,
        test_mode=True,
        clean_filter=clean_filter,
        clean_output=False,
        stop_func=stop_func,
    )

    # The log was streamed
    run_command.tools.subprocess.stream_output.assert_called_once_with(
        label="first",
        popen_process=popen,
        filter_func=mock.ANY,
        stop_func=stop_func,
    )

    # The filter function has the properties we'd expect
    filter_func = run_command.tools.subprocess.stream_output.mock_calls[0].kwargs[
        "filter_func"
    ]
    assert not filter_func.clean_output
    assert filter_func.clean_filter == clean_filter
    assert filter_func.exit_filter.regex.pattern == LogFilter.DEFAULT_EXIT_REGEX


def test_test_mode_failure(run_command, first_app):
    """An app can be streamed in test mode, resulting in test failure."""
    popen = mock.MagicMock()
    popen.returncode = 0
    clean_filter = mock.MagicMock()
    stop_func = mock.MagicMock()
    run_command.tools.subprocess.stream_output = mock.MagicMock()

    # Mock effect of streaming the app resulting in a test suite success
    def mock_stream_output(label, popen_process, filter_func, **kwargs):
        filter_func.returncode = 1

    run_command.tools.subprocess.stream_output.side_effect = mock_stream_output

    # Streaming the app logs causes a test suite failure
    with pytest.raises(BriefcaseTestSuiteFailure):
        run_command._stream_app_logs(
            first_app,
            popen=popen,
            test_mode=True,
            clean_filter=clean_filter,
            clean_output=False,
            stop_func=stop_func,
        )

    # The log was streamed
    run_command.tools.subprocess.stream_output.assert_called_once_with(
        label="first",
        popen_process=popen,
        filter_func=mock.ANY,
        stop_func=stop_func,
    )

    # The filter function has the properties we'd expect
    filter_func = run_command.tools.subprocess.stream_output.mock_calls[0].kwargs[
        "filter_func"
    ]
    assert not filter_func.clean_output
    assert filter_func.clean_filter == clean_filter
    assert filter_func.exit_filter.regex.pattern == LogFilter.DEFAULT_EXIT_REGEX


def test_test_mode_no_result(run_command, first_app):
    """An app can be streamed in test mode, but with no test result being found."""
    popen = mock.MagicMock()
    popen.returncode = 0
    clean_filter = mock.MagicMock()
    stop_func = mock.MagicMock()
    run_command.tools.subprocess.stream_output = mock.MagicMock()

    # Mock effect of streaming the app but not finding a test result.
    def mock_stream_output(label, popen_process, filter_func, **kwargs):
        filter_func.success = None

    run_command.tools.subprocess.stream_output.side_effect = mock_stream_output

    # Stream the app logs raises an error
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Test suite didn't report a result.",
    ):
        run_command._stream_app_logs(
            first_app,
            popen=popen,
            test_mode=True,
            clean_filter=clean_filter,
            clean_output=False,
            stop_func=stop_func,
        )

    # The log was streamed
    run_command.tools.subprocess.stream_output.assert_called_once_with(
        label="first",
        popen_process=popen,
        filter_func=mock.ANY,
        stop_func=stop_func,
    )

    # The filter function has the properties we'd expect
    filter_func = run_command.tools.subprocess.stream_output.mock_calls[0].kwargs[
        "filter_func"
    ]
    assert not filter_func.clean_output
    assert filter_func.clean_filter == clean_filter
    assert filter_func.exit_filter.regex.pattern == LogFilter.DEFAULT_EXIT_REGEX


def test_test_mode_custom_filters(run_command, first_app):
    """An app can define custom success/failure regexes."""
    popen = mock.MagicMock()
    popen.returncode = 0
    clean_filter = mock.MagicMock()
    stop_func = mock.MagicMock()
    run_command.tools.subprocess.stream_output = mock.MagicMock()

    # Mock effect of streaming the app resulting in a test suite success
    def mock_stream_output(label, popen_process, filter_func, **kwargs):
        filter_func.returncode = 0

    run_command.tools.subprocess.stream_output.side_effect = mock_stream_output

    first_app.exit_regex = "THIS IS WHAT SUCCESS LOOKS LIKE"

    # Stream the app logs
    run_command._stream_app_logs(
        first_app,
        popen=popen,
        test_mode=True,
        clean_filter=clean_filter,
        clean_output=False,
        stop_func=stop_func,
    )

    # The log was streamed
    run_command.tools.subprocess.stream_output.assert_called_once_with(
        label="first",
        popen_process=popen,
        filter_func=mock.ANY,
        stop_func=stop_func,
    )

    # The filter function has the properties we'd expect
    filter_func = run_command.tools.subprocess.stream_output.mock_calls[0].kwargs[
        "filter_func"
    ]
    assert not filter_func.clean_output
    assert filter_func.clean_filter == clean_filter
    assert filter_func.exit_filter.regex.pattern == "THIS IS WHAT SUCCESS LOOKS LIKE"


def test_run_app_failure(run_command, first_app):
    """If an app exits with a bad exit code, that is raised as an error."""
    popen = mock.MagicMock()
    popen.returncode = 1
    clean_filter = mock.MagicMock()
    stop_func = mock.MagicMock()
    run_command.tools.subprocess.stream_output = mock.MagicMock()

    # Streaming the app logs raises an error
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Problem running app first.",
    ):
        run_command._stream_app_logs(
            first_app,
            popen=popen,
            test_mode=False,
            clean_filter=clean_filter,
            clean_output=False,
            stop_func=stop_func,
        )

    # The log was streamed
    run_command.tools.subprocess.stream_output.assert_called_once_with(
        label="first",
        popen_process=popen,
        filter_func=mock.ANY,
        stop_func=stop_func,
    )

    # The filter function has the properties we'd expect
    filter_func = run_command.tools.subprocess.stream_output.mock_calls[0].kwargs[
        "filter_func"
    ]
    assert not filter_func.clean_output
    assert filter_func.clean_filter == clean_filter
    assert filter_func.exit_filter.regex.pattern == LogFilter.DEFAULT_EXIT_REGEX


def test_run_app_log_stream_stream_failure(run_command, first_app):
    """If a log stream returns an error code, the log filter requires it is ignored."""
    popen = mock.MagicMock()
    popen.returncode = 1
    clean_filter = mock.MagicMock()
    stop_func = mock.MagicMock()
    run_command.tools.subprocess.stream_output = mock.MagicMock()

    # The log stream raises an error code, but that doesn't raise an error
    run_command._stream_app_logs(
        first_app,
        popen=popen,
        test_mode=False,
        clean_filter=clean_filter,
        clean_output=False,
        stop_func=stop_func,
        log_stream=True,
    )

    # The log was streamed
    run_command.tools.subprocess.stream_output.assert_called_once_with(
        label="log stream",
        popen_process=popen,
        filter_func=mock.ANY,
        stop_func=stop_func,
    )

    # The filter function has the properties we'd expect
    filter_func = run_command.tools.subprocess.stream_output.mock_calls[0].kwargs[
        "filter_func"
    ]
    assert not filter_func.clean_output
    assert filter_func.clean_filter == clean_filter
    assert filter_func.exit_filter.regex.pattern == LogFilter.DEFAULT_EXIT_REGEX


def test_run_app_log_stream_success(run_command, first_app):
    """If a log streamed app returns successfully, no problem is surfaced."""
    popen = mock.MagicMock()
    popen.returncode = 0
    clean_filter = mock.MagicMock()
    stop_func = mock.MagicMock()
    run_command.tools.subprocess.stream_output = mock.MagicMock()

    # Mock effect of streaming the app and returning successfully
    def mock_stream_output(label, popen_process, filter_func, **kwargs):
        filter_func.returncode = 0

    run_command.tools.subprocess.stream_output.side_effect = mock_stream_output

    # Stream the app logs returns success
    run_command._stream_app_logs(
        first_app,
        popen=popen,
        test_mode=False,
        clean_filter=clean_filter,
        clean_output=False,
        stop_func=stop_func,
        log_stream=True,
    )

    # The log was streamed
    run_command.tools.subprocess.stream_output.assert_called_once_with(
        label="log stream",
        popen_process=popen,
        filter_func=mock.ANY,
        stop_func=stop_func,
    )

    # The filter function has the properties we'd expect
    filter_func = run_command.tools.subprocess.stream_output.mock_calls[0].kwargs[
        "filter_func"
    ]
    assert not filter_func.clean_output
    assert filter_func.clean_filter == clean_filter
    assert filter_func.exit_filter.regex.pattern == LogFilter.DEFAULT_EXIT_REGEX


def test_run_app_log_stream_failure(run_command, first_app):
    """If a log streamed app exits with a failure code, the error is surfaced."""
    popen = mock.MagicMock()
    popen.returncode = 0
    clean_filter = mock.MagicMock()
    stop_func = mock.MagicMock()
    run_command.tools.subprocess.stream_output = mock.MagicMock()

    # Mock effect of streaming the app resulting in a test suite success
    def mock_stream_output(label, popen_process, filter_func, **kwargs):
        filter_func.returncode = 1

    run_command.tools.subprocess.stream_output.side_effect = mock_stream_output

    # Streaming the app logs causes a test suite failure
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Problem running app first.",
    ):
        run_command._stream_app_logs(
            first_app,
            popen=popen,
            test_mode=False,
            clean_filter=clean_filter,
            clean_output=False,
            stop_func=stop_func,
            log_stream=True,
        )

    # The log was streamed
    run_command.tools.subprocess.stream_output.assert_called_once_with(
        label="log stream",
        popen_process=popen,
        filter_func=mock.ANY,
        stop_func=stop_func,
    )

    # The filter function has the properties we'd expect
    filter_func = run_command.tools.subprocess.stream_output.mock_calls[0].kwargs[
        "filter_func"
    ]
    assert not filter_func.clean_output
    assert filter_func.clean_filter == clean_filter
    assert filter_func.exit_filter.regex.pattern == LogFilter.DEFAULT_EXIT_REGEX


def test_run_app_log_stream_no_result(run_command, first_app):
    """If a log streamed app returns no result, no problem is surfaced."""
    popen = mock.MagicMock()
    popen.returncode = 0
    clean_filter = mock.MagicMock()
    stop_func = mock.MagicMock()
    run_command.tools.subprocess.stream_output = mock.MagicMock()

    # Mock effect of streaming the app but not finding a test result.
    def mock_stream_output(label, popen_process, filter_func, **kwargs):
        filter_func.returncode = None

    run_command.tools.subprocess.stream_output.side_effect = mock_stream_output

    # Streaming the app logs doesn't raise an error
    run_command._stream_app_logs(
        first_app,
        popen=popen,
        test_mode=False,
        clean_filter=clean_filter,
        clean_output=False,
        stop_func=stop_func,
        log_stream=True,
    )

    # The log was streamed
    run_command.tools.subprocess.stream_output.assert_called_once_with(
        label="log stream",
        popen_process=popen,
        filter_func=mock.ANY,
        stop_func=stop_func,
    )

    # The filter function has the properties we'd expect
    filter_func = run_command.tools.subprocess.stream_output.mock_calls[0].kwargs[
        "filter_func"
    ]
    assert not filter_func.clean_output
    assert filter_func.clean_filter == clean_filter
    assert filter_func.exit_filter.regex.pattern == LogFilter.DEFAULT_EXIT_REGEX


def test_run_app_ctrl_c(run_command, first_app):
    """An app can have its logs streamed, but be interrupted."""
    popen = mock.MagicMock()
    popen.returncode = 0
    clean_filter = mock.MagicMock()
    stop_func = mock.MagicMock()
    run_command.tools.subprocess.stream_output = mock.MagicMock()

    # Mock a CTRL-C
    run_command.tools.subprocess.stream_output.side_effect = KeyboardInterrupt

    # Stream the app logs
    run_command._stream_app_logs(
        first_app,
        popen=popen,
        test_mode=False,
        clean_filter=clean_filter,
        clean_output=False,
        stop_func=stop_func,
    )

    # The log was streamed
    run_command.tools.subprocess.stream_output.assert_called_once_with(
        label="first",
        popen_process=popen,
        filter_func=mock.ANY,
        stop_func=stop_func,
    )

    # The filter function has the properties we'd expect
    filter_func = run_command.tools.subprocess.stream_output.mock_calls[0].kwargs[
        "filter_func"
    ]
    assert not filter_func.clean_output
    assert filter_func.clean_filter == clean_filter
    assert filter_func.exit_filter.regex.pattern == LogFilter.DEFAULT_EXIT_REGEX



import subprocess
import sys
from pathlib import Path

import pytest

from briefcase.exceptions import BriefcaseCommandError, InvalidDeviceError


def test_simple_command(mock_tools, adb, tmp_path):
    """ADB.run() invokes adb with the provided arguments."""
    adb.run("example", "command")

    # Check that adb was invoked with the expected commands
    mock_tools.subprocess.check_output.assert_called_once_with(
        [
            tmp_path
            / f"sdk/platform-tools/adb{'.exe' if sys.platform == 'win32' else ''}",
            "-s",
            "exampleDevice",
            "example",
            "command",
        ],
        quiet=False,
    )


def test_quiet_command(mock_tools, adb, tmp_path):
    """ADB.run() can be invoked in quiet mode."""
    adb.run("example", "command", quiet=True)

    # Check that adb was invoked with the expected commands
    mock_tools.subprocess.check_output.assert_called_once_with(
        [
            tmp_path
            / f"sdk/platform-tools/adb{'.exe' if sys.platform == 'win32' else ''}",
            "-s",
            "exampleDevice",
            "example",
            "command",
        ],
        quiet=True,
    )


@pytest.mark.parametrize(
    "name, exception",
    [
        # When the device is not found, look for a command the user can run to get a
        # list of valid devices.
        ("device-not-found", InvalidDeviceError),
        # Validate that when an arbitrary adb errors, we print the full adb output.
        # This adb output comes from asking it to run a nonexistent adb command.
        ("arbitrary-adb-error-unknown-command", subprocess.CalledProcessError),
    ],
)
def test_error_handling(mock_tools, adb, name, exception, tmp_path):
    """ADB.run() can parse errors returned by adb."""
    # Set up a mock command with a subprocess module that has with sample data loaded.
    adb_samples = Path(__file__).parent / "adb_errors"
    with (adb_samples / (name + ".out")).open("r", encoding="utf-8") as adb_output_file:
        with (adb_samples / (name + ".returncode")).open(
            encoding="utf-8"
        ) as returncode_file:
            mock_tools.subprocess.check_output.side_effect = (
                subprocess.CalledProcessError(
                    returncode=int(returncode_file.read().strip()),
                    cmd=["ignored"],
                    output=adb_output_file.read(),
                )
            )

    # invoke run()
    with pytest.raises(exception):
        adb.run("example", "command")

    # Check that adb was invoked as expected
    mock_tools.subprocess.check_output.assert_called_once_with(
        [
            tmp_path
            / f"sdk/platform-tools/adb{'.exe' if sys.platform == 'win32' else ''}",
            "-s",
            "exampleDevice",
            "example",
            "command",
        ],
        quiet=False,
    )


def test_older_sdk_error(mock_tools, adb):
    """Failure [INSTALL_FAILED_OLDER_SDK] needs to be caught manually."""
    mock_tools.subprocess.check_output.return_value = "\n".join(
        [
            "Performing Push Install",
            "C:/.../app-debug.apk: 1 file pushed, 0 skipped. 5.5 MB/s (33125287 bytes in 5.768s)",
            "         pkg: /data/local/tmp/app-debug.apk",
            "Failure [INSTALL_FAILED_OLDER_SDK]",
        ]
    )
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Your device doesn't meet the minimum SDK requirements of this app",
    ):
        adb.run("example", "command")



from __future__ import annotations

from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseCommandError

from .base import BaseCommand, full_options


class BuildCommand(BaseCommand):
    command = "build"
    description = "Build an app for a target platform."

    def add_options(self, parser):
        self._add_update_options(parser, context_label=" before building")
        self._add_test_options(parser, context_label="Build")

    def build_app(self, app: AppConfig, **options):
        """Build an application.

        :param app: The application to build
        """
        # Default implementation; nothing to build.

    def _build_app(
        self,
        app: AppConfig,
        update: bool,
        update_requirements: bool,
        update_resources: bool,
        update_support: bool,
        update_stub: bool,
        no_update: bool,
        test_mode: bool,
        **options,
    ) -> dict | None:
        """Internal method to invoke a build on a single app. Ensures the app exists,
        and has been updated (if requested) before attempting to issue the actual build
        command.

        :param app: The application to build
        :param update: Should the application be updated before building?
        :param update_requirements: Should the application requirements be updated
            before building?
        :param update_resources: Should the application resources be updated before
            building?
        :param update_support: Should the application support be updated?
        :param update_stub: Should the stub binary be updated?
        :param no_update: Should automated updates be disabled?
        :param test_mode: Is the app being build in test mode?
        """
        if not self.bundle_path(app).exists():
            state = self.create_command(app, test_mode=test_mode, **options)
        elif (
            update  # An explicit update has been requested
            or update_requirements  # An explicit update of requirements has been requested
            or update_resources  # An explicit update of resources has been requested
            or update_support  # An explicit update of app support has been requested
            or update_stub  # An explicit update of the stub binary has been requested
            or (
                test_mode and not no_update
            )  # Test mode, but updates have not been disabled
        ):
            state = self.update_command(
                app,
                update_requirements=update_requirements,
                update_resources=update_resources,
                update_support=update_support,
                update_stub=update_stub,
                test_mode=test_mode,
                **options,
            )
        else:
            state = None

        self.verify_app(app)

        state = self.build_app(app, test_mode=test_mode, **full_options(state, options))

        qualifier = " (test mode)" if test_mode else ""
        self.logger.info(
            f"Built {self.binary_path(app).relative_to(self.base_path)}{qualifier}",
            prefix=app.app_name,
        )
        return state

    def __call__(
        self,
        app: AppConfig | None = None,
        update: bool = False,
        update_requirements: bool = False,
        update_resources: bool = False,
        update_support: bool = False,
        update_stub: bool = False,
        no_update: bool = False,
        test_mode: bool = False,
        **options,
    ) -> dict | None:
        # Has the user requested an invalid set of options?
        # This can't be done with argparse, because it isn't a simple mutually exclusive group.
        if no_update:
            if update:
                raise BriefcaseCommandError(
                    "Cannot specify both --update and --no-update"
                )
            if update_requirements:
                raise BriefcaseCommandError(
                    "Cannot specify both --update-requirements and --no-update"
                )
            if update_resources:
                raise BriefcaseCommandError(
                    "Cannot specify both --update-resources and --no-update"
                )
            if update_support:
                raise BriefcaseCommandError(
                    "Cannot specify both --update-support and --no-update"
                )
            if update_stub:
                raise BriefcaseCommandError(
                    "Cannot specify both --update-stub and --no-update"
                )

        # Confirm host compatibility, that all required tools are available,
        # and that the app configuration is finalized.
        self.finalize(app)

        if app:
            state = self._build_app(
                app,
                update=update,
                update_requirements=update_requirements,
                update_resources=update_resources,
                update_support=update_support,
                update_stub=update_stub,
                no_update=no_update,
                test_mode=test_mode,
                **options,
            )
        else:
            state = None
            for app_name, app in sorted(self.apps.items()):
                state = self._build_app(
                    app,
                    update=update,
                    update_requirements=update_requirements,
                    update_resources=update_resources,
                    update_support=update_support,
                    update_stub=update_stub,
                    no_update=no_update,
                    test_mode=test_mode,
                    **full_options(state, options),
                )

        return state



import os
import sys
from unittest.mock import MagicMock

import pytest

from briefcase.console import Console, Log
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.windows.app import WindowsAppOpenCommand


@pytest.fixture
def open_command(tmp_path, first_app_config):
    command = WindowsAppOpenCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.os = MagicMock(spec_set=os)
    command.tools.subprocess = MagicMock(spec_set=Subprocess)
    return command


@pytest.mark.skipif(sys.platform != "win32", reason="Windows specific test")
def test_open_windows(open_command, first_app_config, tmp_path):
    """On Windows, open invokes `startfile` on the project folder."""
    # Create the project folder to mock a created project.
    open_command.project_path(first_app_config).mkdir(parents=True)

    open_command(first_app_config)

    open_command.tools.os.startfile.assert_called_once_with(
        tmp_path / "base_path/build/first-app/windows/app"
    )
